using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;

using UnityEngine;
using UnityEngine.Events;
using UnityEngine.UI;
using UnityEngine.XR;

using SimpleJSON;

namespace Amaimon {
    public abstract class APlugin {
        public JSONStorable storables = null;

        public APlugin(AActor actor, string partialName) {
            storables = Unity.GetStorableByPartialName(actor.person, partialName);
        }

        public APlugin(JSONStorable s) {
            storables = s;
        }

        public APlugin(string pluginName) {
            storables = Unity.findPluginAnywhere(pluginName);
            if(storables == null)
                SuperController.LogError($"Couldn't find plugin {pluginName} in whole scene!");
        }

        public abstract void init();

        public bool isEnabled {
            set {
                storables?.SetBoolParamValue("enabled", value);
            }
            get {
                return storables.GetBoolJSONParam("enabled").val;
            }
        }
    }

    public abstract class NamedProperty<T>
    {
        protected APlugin _plugin;
        protected string _name;

        public NamedProperty(APlugin plugin, string name) {
            _plugin = plugin;
            _name = name;
        }
        public abstract T get();
        public abstract void set(T value);
    }

    public class ActionProperty : NamedProperty<string> {
        public ActionProperty(APlugin plugin, string name) : base(plugin, name) {}

        public override string get() {
            return null;
        }

        public override void set(string value) {}

        public void call() {
            _plugin.storables.CallAction(_name);
        }
    }

    public class BoolProperty : NamedProperty<bool> {
        public BoolProperty(APlugin plugin, string name) : base(plugin, name) {}

        public override bool get() {
            return _plugin.storables.GetBoolJSONParam(_name).val;
        }

        public override void set(bool value) {
            _plugin.storables.SetBoolParamValue(_name, value);
        }
    }

    public class StringProperty : NamedProperty<string> {
        public StringProperty(APlugin plugin, string name) : base(plugin, name) {}

        public override string get() {
            return _plugin.storables.GetStringJSONParam(_name).val;
        }

        public override void set(string value) {
            _plugin.storables.SetStringParamValue(_name, value);
        }
    }

    public class FloatProperty : NamedProperty<float> {
        public FloatProperty(APlugin plugin, string name) : base(plugin, name) {}

        public override float get() {

            return _plugin.storables.GetFloatJSONParam(_name).val;
        }

        public override void set(float value) {
            // bool isNull = _plugin.storables == null;
            // Log.d($"{_name} plugin storables null? {isNull}");
            _plugin.storables.SetFloatParamValue(_name, value);
        }
    }

    // JSONStorableStringChooser
    public class StringChooserProperty : NamedProperty<string> {
        public StringChooserProperty(APlugin plugin, string name) : base(plugin, name) {}

        public override string get() {
            return _plugin.storables.GetStringChooserJSONParam(_name).val;
        }

        public override void set(string value) {
            JSONStorableStringChooser test = _plugin.storables.GetStringChooserJSONParam(_name);
            //Log.d($"setting chooser {value}:{test}={test.val}");
            _plugin.storables.GetStringChooserJSONParam(_name).val = value;
        }
    }


}