using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System;

namespace Amaimon {
    //dodać kanały na interpolatory
    // 1. kanał oddechu
    // 2. kanał mimiki + baselinów
    // 3. kanał ruchów ust, blockedBy 2.
    // 4. kanał ruchów ciała
    // 5. kanał ruchu głowy
    //
    // queue - odpali po następnym
    // maybe - odpali, jak nic się nie dzieje na tym kanale
    // preempt - zatrzymuje wszystko na tym kanale, puszcza to
    // blockedBy x - nie odpala nowych zadań, jeżeli na X jest alive


    public class SmartChannelJob {
        public float timer = 0;
        public Func<float> job = null;

        public SmartChannelJob(Func<float> a) {
            job = a;
        }
    }

    public class SmartChannel : InterpolableChannel<LifecycleInterpolable> {
        LifecycleInterpolable mainInterpolable = null;
        List<SmartChannelJob> jobs = new List<SmartChannelJob>();
        public SmartChannelJob mainJob = null;

        public void overrideMainInterpolable(LifecycleInterpolable i, bool pre, float duration) {
            purge();
            //Log.d($"Overriding main witrh {duration}, main={mainInterpolable}");
            if(pre || mainInterpolable == null || mainInterpolable.stoppedOrStopping() == true) {
                //Log.d("Override success");
                mainInterpolable?.stop();
                if(mainJob != null)
                    mainJob.timer = duration;
                mainInterpolable = i;
                preempt(i);   
            }
        }

        public void stopOverrideMainInterpolable() {
            // Log.d($"Stop Overriding main");
            // mainInterpolable?.stop();
            // purge();
            // mainJob.timer = mainJob.job(0.01f);
        }

        public void addIdleJob(Func<float> j) {
            purge();
            var job = new SmartChannelJob(j);
            jobs.Add(job);
        }

        public void setMainJob(Func<float> j, bool preempt = true) {
            mainJob = null;
            mainInterpolable?.stop();
            var job = new SmartChannelJob(j);
            mainJob = job;
        }

        public void noIdle() {
            jobs.Clear();
        }

        override public void stop() {
            mainJob = null;
            jobs.Clear();
            mainInterpolable?.stop();
            base.stop();
        }

        override public void onTick(float deltaT) {
            base.onTick(deltaT);
            if(mainJob != null) {
                mainJob.timer-=deltaT;
            }

            //Log.d($"main timer={mainJob.timer}");

            if(mainJob?.timer <= 0f) {
                mainJob.timer = mainJob.job();
            }
            else if(mainInterpolable == null ||  mainInterpolable?.stoppedOrStopping()==true) {
                jobs.ForEach(job => {
                    job.timer -= deltaT;
                    if(job.timer <= 0f) {
                        try {
                            job.timer = job.job();
                        } catch (Exception ex) {

                        }
                    }
                });
            }
        }
    }

    public class InterpolableChannel<T> : InterpolablePackGen<T> where T : LifecycleInterpolable {
        InterpolableChannel<T> blockedBy = null;
        public void queue(T i) {
            // dodać i uruchomić, jak wszystkie przed nim przestaną być live, czyli przestawić mu isAlive na false
            add(i);
        }

        public void run(T i) {
            // run in parallel with other existing in the cahannel
            add(i);
            i.runNow();
        }
        public void maybe(T i) {
            // run only if nothing else is running on this channel, otherwise - ignore
            // TODO - potrzebna magia, która pozwoli wymusić przerwanie baseline
            // może kanał baseline musi być blokowany przez kanał expressions?
            purge();
            if(stoppedOrStopping()) {
                // Log.d("maybe - adding new interpolable");
                // Log.d($"maybe {i}");

                // if(typeof(T)==typeof(ExpressionMorph)) {
                //     AMorph am = i as AMorph;
                //     Log.d($"maybe - Applying morph {am.getName()}");
                // }


                add(i);
                i.runNow();
            }
            else {
                //Log.d("maybe - denied!");
            }
        }
        public void preempt(T ii) {
            // all must stop, this must run
            // Log.d("preempting new interpolable");

            // if(typeof(T)==typeof(ExpressionMorph)) {
            //     AMorph am = ii as AMorph;
            //     Log.d($"maybe - Applying morph {am.getName()}");
            // }


            list.Where(i => i.getIsTicking()).ForEach(i => i.stop());
            //Log.d("Preempt:");
            add(ii);
            ii.runNow();
        }

        virtual public void onTick(float deltaT) {
            // robimy onTick na wszystkich żywych
            base.onTick(deltaT);

            var everythingLiveStoppedOrStopping = list.Where(element => !element.waitingToBeRun).All(element => element.stoppedOrStopping());

            if(everythingLiveStoppedOrStopping) {
                // var waiting = list.Where(i => i.waitingToBeRun).Count;
                
                // if(waiting>0) {
                //     waiting.First().runNow();
                // }
                list.FirstOrDefault(i => i.waitingToBeRun)?.runNow();
            }
            // jeżeli na szczycie jest jakiś oczekujący, a nic już nie żyje, to odpalamy go, o ile blockedBy nie jest alive
            // patrzymy, czy od góry są jakieś oczekujące. Jeżeli tak, to odpalamy je, o ile blocker.getIsRunning() == false

        }
    }

    public class LifecycleInterpolableChannel : InterpolableChannel<LifecycleInterpolable> {

    }

    public class ExpressionInterpolableChannel : InterpolableChannel<ExpressionMorph> {
        
    }
}