using UnityEngine;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;

namespace Amaimon {
    ///////////////////////////////////////////////////
    // Regular morph that has a duration and max value
    // used for expressions
    ///////////////////////////////////////////////////

    public class AMorph : SmoothEnvelopeInterpolator {
        DAZMorph _morph; 
        string _name;

        public AMorph(AActor person, string name, float target, float inTime, float outTime, float sustain = 0) : base(target, inTime, outTime, sustain) {
            _morph = person.getMorph(name);
            if(_morph == null)
                Log.d($"Couldn't find morph {name} on {person.person.name}");
            else
                _default = 0;//_morph.morphValue;
            _name = name;
        }

        public string getName() {
            return _name;
        }

        override public void apply(float v) {
            //Log.d($"Morf idzie {v}");
            _morph.morphValue=v;
        }

        // override public void onTick(float deltaTimeS) {
        //     Log.d("on tick");
        //     base.onTick(deltaTimeS);
        // }

        // public override bool getIsTicking() {
        //     var v = !waitingToBeRun && _phase > 0;
        //     Log.d($"getIsTicking waiting={waitingToBeRun} phase={_phase}");
        //     return !waitingToBeRun && _phase > 0;
        // }

        public bool isSameMorph(AMorph other) {
            return this._morph == other._morph;
        }

        public override void runNow() {
            _reached = _morph.morphValue;
            _phase = 1;
            waitingToBeRun = false;
        }
    }


    public class ExpressionMorph : AMorph {
        public ExpressionMorph(AActor person, string name, float target, float inTime, float outTime, float sustain = 0) : base(person, name, target, inTime, outTime) {
            _sustain = sustain;
            _default = 0;
        }
    }



    ///////////////////////////////////////////////////
    // On/off morph that is a switch
    // used for mood morphs (baselines)
    ///////////////////////////////////////////////////

    public class OnOffMorph : OnOffInterpolator {
        public DAZMorph _morph; 

        public OnOffMorph(AActor person, string name, float onVal, float offVal, float inTime, float outTime, Func <float, float> ac = null) : base(onVal, offVal, inTime, outTime, ac) {
            _morph = person.getMorph(name);
            if(_morph == null)
                Log.d($"Couldn't find morph {name} on {person.person.name}");
        }

        public OnOffMorph(AActor person, string name, float onVal, float inTime, float outTime, Func <float, float> ac = null) : base(onVal, 0, inTime, outTime, ac) {
            _morph = person.getMorph(name);
            if(_morph == null)
                Log.d($"Couldn't find morph {name} on {person.person.name}");
        }

        public bool isSameMorph(OnOffMorph other) {
            return this._morph == other._morph;
        }

        override public void apply(float v) {
            _morph.morphValue=v;
        }
    }   
}