using UnityEngine;
using System;

namespace Amaimon {
    /*
            // chwilowa siła -150 do 150 daje spoko efekt
            // moment -60 do 60 - spory, celowy obrót, coś jakby Marylin Monroe szła sobie...
        */

    class ShakyRotationSpringInterpolator : NoisyInterpolator {
        FreeControllerV3 contr;

        public ShakyRotationSpringInterpolator(AActor female, string controlName, float amount, float speed) : base(amount, speed) {
            // amount 0-500
            // speed 0-10
            contr = female.getController(controlName);
            _default = contr.RBHoldRotationSpring;
            _phase = 2;
        }

        override public void apply(float value) {
            contr.RBHoldRotationSpring = value;
        }
    }

    class ShakyPositionSpringInterpolator : NoisyInterpolator {
        FreeControllerV3 contr;

        public ShakyPositionSpringInterpolator(AActor female, string controlName, float amount, float speed) : base(amount, speed) {
            // amount 0-500
            // speed 0-10
            contr = female.getController(controlName);
            _default = contr.RBHoldPositionSpring;
            _phase = 2;
        }

        override public void apply(float value) {
            contr.RBHoldPositionSpring = value;
        }
    }



    class ForceCyclicInterpolator : SinusaidalInterpolator {
        FreeControllerV3 contr;
        Vector3 _axis;

        public ForceCyclicInterpolator(AActor female, string controlName, Vector3 axis, float amplitude, float perdiodSecs, float sustain=100000) : base(amplitude, perdiodSecs, sustain) {
            contr = female.getController(controlName);
            _axis = axis;
            _phase = 2;
        }

        override public void apply(float value) {
            contr.followWhenOffRB.AddForce(_axis * value);
        }
    }

    class TorqueCyclicInterpolator : SinusaidalInterpolator {
        FreeControllerV3 contr;
        Vector3 _axis;

        public TorqueCyclicInterpolator(AActor female, string controlName, Vector3 axis, float amplitude, float perdiodSecs, float sustainPeriods=100000) : base(amplitude, perdiodSecs, sustainPeriods) {
            contr = female.getController(controlName);
            _axis = axis;
            _phase = 2;
        }

        override public void apply(float value) {
            contr.followWhenOffRB.AddRelativeTorque(_axis * value);
        }
    }

    class HoolaCycle : TorqueCyclicInterpolator {
        FreeControllerV3 contr;
        public HoolaCycle(AActor actor, string controlName, float amplitude, float perdiodSecs, float sustainPeriods=100000) : base(actor, controlName, Vector3.up, amplitude, perdiodSecs, sustainPeriods) {}
    }

    class HumpCycle : TorqueCyclicInterpolator {
        FreeControllerV3 contr;
        public HumpCycle(AActor actor, string controlName, float amplitude, float perdiodSecs, float sustain=100000) : base(actor, controlName, Vector3.right, amplitude, perdiodSecs, sustain) {}
    }

    class SwingCycle : TorqueCyclicInterpolator {
        FreeControllerV3 contr;
        public SwingCycle(AActor actor, string controlName, float amplitude, float perdiodSecs, float sustain=100000) : base(actor, controlName, Vector3.forward, amplitude, perdiodSecs, sustain) {}
    }
}