using System;
using System.Linq;
using System.Collections;
using System.Collections.Generic;

using UnityEngine;

namespace Amaimon {
    public class ProbableEntry<T>{
        public T entry;
        public float probability;
        public float priority;

        public ProbableEntry(T e, float p) {
            entry = e;
            probability = p;
        }

        public override string ToString () {
            return $"{entry} - {priority}";
        }
    }

    public class RandomTable<T> {
        public List<ProbableEntry<T>> entries = new List<ProbableEntry<T>>();
        float lastProb = 0;
        private int next = 0;

        public static RandomTable<T> create() {
            return new RandomTable<T>();
        }

        public bool isEmpty() {
            return entries.Count == 0;
        }

        public void clear() {
            entries.Clear();
        }

        public void prioritiesToChance() {
            lastProb = 0;
            entries.ForEach(entry => {
                lastProb+=entry.priority;
                entry.probability = lastProb;
            });
        }

        public RandomTable<T> add(T entry, float p) {
            lastProb+=p;
            entries.Add(new ProbableEntry<T>(entry, lastProb));
            return this;
        }
        
        public RandomTable<T> addIfNotNull(T entry, float p) {
            if(entry != null) {
                add(entry, p);
            }
            return this;
        }

        public void normalize() {
            var max = entries.Max(it => it.priority);
            //Log.d($"max={max}");
            if(max != 0) {
                entries.ForEach(entry => entry.priority = entry.priority / max);
            }
        }

        public void dump() {
            Log.d($"==========================");
            entries.ForEach(entry => {
                Log.d($"{entry}");
            });
        }

        public T getRandom() { 
            var random = UnityEngine.Random.Range(0, lastProb);
            //Log.d($"Froma a table of {lastProb} choosing {random}");
            try {
                return entries.First(entry => entry.probability > random).entry;
            } catch (Exception ex) {
                return default(T);
            }
        }
    }
}