using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;

using UnityEngine;
using UnityEngine.Events;
using UnityEngine.UI;
using UnityEngine.XR;

using SimpleJSON;

namespace Amaimon {

    public class Segment
    {
        public bool enabled = true;
        public string name;
        public DAZBone bone;
        public float x; //ui slider for the segment's x rotation
        public float y; //ui slider for the segment's y rotation
        public float z; //ui slider for the segment's z rotation
        public Segment(string s)
        {
            name = s;
        }
        public void UpdateRotation(float f)
        {
            if (enabled) bone.baseJointRotation = new Vector3(x, y, z);
        }
        public void ResetRotation()
        {
            x = 0;
            y = 0;
            z = 0;
            UpdateRotation(0f);
        }
        public void ResetBoneRotation()
        {
            bone.baseJointRotation = Vector3.zero;
        }
    }
    public class Finger
    {
        public string name;
        public Segment[] segment = new Segment[3]
        {
            new Segment("1"),
            new Segment("2"),
            new Segment("3")
        };
        //public UIDynamicButton UIToggleButton;
        //public bool UIExpanded = false;
        public Finger(string s)
        {
            name = s;
        }
        
        

        public void openClose(float power, bool isLeft) {
            // zaciskanie dłoni: z2 i z3 w palcach na identyczny minus, poza środkowym. środkowy z2 przy -40 ustawić na -50
            // dobry setting: z1 -15 z2 -40, z3 -20, tylko środkowy inaczej z2 -50
            // do tego kciuk: z1 -50, y3 40

            // przeprosty palców, rozczapirzona dłoń: +14 z1, z2, z3, kciuk z1 +14 y1, y3 -14

            // czyli - gdy 0, musimy ustawić na 14
            var coeff = (isLeft) ? -1 : 1;

            if(name=="Thumb") {
                float z1power = 14 - power*50;
                float y1power = -14 + power*14;
                float y3power = -14 + power*40;
                segment[0].z = coeff*z1power;
                segment[0].y = coeff*y1power;
                segment[2].y = coeff*y3power;
            }
            else {
                float z1power = 14 - power*15;
                float z2power = 14 - power*30;
                if(name=="Index")
                    z2power = 14 - power*40;
                float z3power = 14 - power*40;
                segment[0].z = coeff*z1power;
                segment[1].z = coeff*z2power;
                segment[2].z = coeff*z3power;
                //Log.d($"{name} {segment[0].z} {segment[1].z} {segment[2].z}");
            }
            UpdateRotations();
        }
        
        public void UpdateRotations()
        {
            foreach (Segment s in segment) s.UpdateRotation(0f);
        }
        public void ResetRotations()
        {
            foreach (Segment s in segment) s.ResetRotation();
        }
        public void ResetBoneRotations()
        {
            foreach (Segment s in segment) s.ResetBoneRotation();
        }
    }
    public class Hand
    {
        public bool IsRight = true;
        public string name = "Right"; //right hand by default

        public FreeControllerV3 controller;

        public Finger[] finger = new Finger[5]
        {
            new Finger("Thumb"),
            new Finger("Index"),
            new Finger("Middle"),
            new Finger("Ring"),
            new Finger("Pinky")
        };
        public Segment[] carpal = new Segment[2]
            {
                new Segment("1"),
                new Segment("2")
            };
        //public UIDynamicButton UICarpalToggle;
        //public bool UIExpandedCarpals = false;
        //public UIDynamicButton SavePoseButton;
        //public UIDynamicButton LoadPoseButton;
        //public UIDynamicButton ResetButton;
        //public JSONStorableString PoseNameField;
        public string PoseOutputFormat;

        public Hand(bool b, AActor a)
        {
            IsRight = b;
            if (b) {
                controller = a.rightHandController;
                name = "Right";
            }
            else {
                controller = a.leftHandController;
                name = "Left";
            }
        }
        public void SetEnabled(bool b)
        {
            foreach (Finger f in finger)
            {
                foreach (Segment s in f.segment) s.enabled = b;
            }
            foreach (Segment s in carpal) s.enabled = b;
        }

        public void openClose(float power) {
            foreach (Finger f in finger) f.openClose(power, name=="Left");
            //foreach (Segment s in carpal) s.UpdateRotation(0f);
        }

        public void overstretch() {
            foreach (Finger f in finger) f.openClose(2f, !IsRight);
        }
        public void flat() {
            foreach (Finger f in finger) f.openClose(0.82f, !IsRight);
        }
        public void relaxed() {
            foreach (Finger f in finger) f.openClose(0.41f, !IsRight);
        }
        public void squeezy() {
            foreach (Finger f in finger) f.openClose(0.32f, !IsRight);
        }

        public void pinch() {
            foreach (Finger f in finger) f.openClose(-0.18f, !IsRight);
        }

        public void grab() {
            foreach (Finger f in finger) f.openClose(-0.42f, !IsRight);
        }

        public void fist() {
            foreach (Finger f in finger) f.openClose(-1f, !IsRight);
        }



        public void point() {
            finger[0].openClose(0.82f, !IsRight);
            finger[1].openClose(1.5f, !IsRight);    
            finger[2].openClose(-1f, !IsRight);    
            finger[3].openClose(-1f, !IsRight);    
            finger[4].openClose(-1f, !IsRight);    
        }

        public void upYours() {
            finger[0].openClose(-0.5f, !IsRight);
            finger[1].openClose(-1f, !IsRight);    
            finger[2].openClose(1.8f, !IsRight);    
            finger[3].openClose(-1f, !IsRight);    
            finger[4].openClose(-1f, !IsRight);    
        }

        public void aveSatan() {
            finger[0].openClose(-0.8f, !IsRight);
            finger[1].openClose(1.8f, !IsRight);    
            finger[2].openClose(-1f, !IsRight);    
            finger[3].openClose(-1f, !IsRight);    
            finger[4].openClose(1.8f, !IsRight);    
        }

        public void heart() {
            finger[0].openClose(2f, !IsRight);
            finger[1].openClose(-1f, !IsRight);    
            finger[2].openClose(-1f, !IsRight);    
            finger[3].openClose(-1f, !IsRight);    
            finger[4].openClose(2f, !IsRight);    
        }

        public void ok() {
            finger[0].openClose(2f, !IsRight);
            finger[1].openClose(-1f, !IsRight);    
            finger[2].openClose(-1f, !IsRight);    
            finger[3].openClose(-1f, !IsRight);    
            finger[4].openClose(-1f, !IsRight);    
        }

        public void UpdateRotations()
        {
            foreach (Finger f in finger) f.UpdateRotations();
            foreach (Segment s in carpal) s.UpdateRotation(0f);
        }
        public void ResetRotations()
        {
            foreach (Finger f in finger) f.ResetRotations();
            foreach (Segment s in carpal) s.ResetRotation();
        }
        public void ResetBoneRotations()
        {
            foreach (Finger f in finger) f.ResetBoneRotations();
            foreach (Segment s in carpal) s.ResetBoneRotation();
        }
        public void UpdateSlidersFromBones()
        {
            foreach (Segment s in GetSegments())
            {
                s.x = s.bone.baseJointRotation.x;
                s.y = s.bone.baseJointRotation.y;
                s.z = s.bone.baseJointRotation.z;
            }
        }
        public List<Segment> GetSegments()
        {
            List<Segment> l = new List<Segment>();
            foreach (Finger f in finger)
            {
                foreach (Segment s in f.segment) l.Add(s);
            }
            foreach (Segment s in carpal) l.Add(s);
            return l;
        }
    }





    public partial class AActor : AAtom {
        RandomTable<PriorityEntry> touchTable = new RandomTable<PriorityEntry>();


        public void handsOff() {
            touchTable.clear();
        }

        public void addHandsOn(AAtom whom) {
            whom.getTouchTargets().ForEach(target => {
                touchTable.add(new PriorityEntry(target),0);

            });
        }

        public float prioritizeHands(bool isRight) {

            // zrobić cooldown per ręka, nie per część ciała!
            touchTable.entries.ForEach(i => {

                i.priority = 0;

                if(i.entry.body != null) {
                    Vector3 entryPos = i.entry.body.transform.position;

                    if(i.entry.body.niceName().Contains("hand")) {
                        // TODO - reach for hand if within 30cm                        
                    }
                    else {

                        var dist = (isRight) ? Vector3.Distance(rShoulderRb.transform.position, entryPos) : Vector3.Distance(lShoulderRb.transform.position, entryPos);

                        if(dist < 0.2f) {
                            i.priority += 2.5f-dist;
                        }
                        else if(dist < 0.5f) {
                            i.priority += 2.0f-dist;
                        }
                        else if(dist < 0.8f) {
                            i.priority += 1.5f-dist;
                        }

                        if(i.entry.body.niceName().Contains("gen")) {
                            i.priority += arousal * dist;
                        }
                    }
                }                        
            });

            touchTable.prioritiesToChance();

            // Log.d($"{person.name}");
            //touchTable.dump();

            var randomTarget = touchTable.getRandom();

            if(randomTarget == null) {
                Log.d($"{person.name} rightHand={isRight} got null targert!");
                // touchTable.dump();
                stopReeachingWithArm(isRight);
                return 0;
            }
            else if(randomTarget.body != null) {
                var pause = getScaredRelaxedInterval() * 4;
                touchInterest(randomTarget.body.transform, randomTarget.body.niceName(), pause, isRight);
                return pause;
            }
            else {
                Log.d($"{person.name} error - got target but nor rigidbody");
                stopReeachingWithArm(isRight);
                return 5f;
            }
        }

        void touchInterest(Transform target, string targetName, float touchDuration, bool isRight) {

            var destinationPos = target.position;
            var destinationNor = target.rotation;

            var rdist = Vector3.Distance(rShoulderRb.transform.position, destinationPos);
            var ldist = Vector3.Distance(lShoulderRb.transform.position, destinationPos);

            if(rdist >= 0.7f) {
                stopReeachingWithArm(true);
            }

            if(ldist >= 0.7f) {
                stopReeachingWithArm(false);
            }

            FreeControllerV3 activeHand;
            Hand activePalm;

            if(rdist <= ldist) {
                activeHand = prepareRightArmForReach();
                activePalm = rightHand;
            }
            else {
                activeHand = prepareLeftArmForReach();
                activePalm = leftHand;
            }

            Quaternion handRotation;

            var directionVector = target.position - activeHand.transform.position;

            if(targetName.Contains("rhand")) {
                //                                       blue          red
                handRotation = Quaternion.LookRotation(directionVector, target.up);
                activePalm.overstretch();
            }
            else if(targetName.Contains("lhand")) {
                //                                       blue          red
                handRotation = Quaternion.LookRotation(directionVector, target.up);
                activePalm.overstretch();
            }
            else if(targetName.Contains("head")) {
                if(isRight)
                    handRotation = Quaternion.LookRotation(target.transform.up, -target.transform.right);
                else
                    handRotation = Quaternion.LookRotation(target.transform.up, target.transform.right);
                activePalm.relaxed();
            }
            else if(targetName.Contains("abdomen")) {
                if(isRight)
                    handRotation = Quaternion.LookRotation(-target.transform.right, -target.transform.forward);
                else
                    handRotation = Quaternion.LookRotation(-target.transform.forward, -target.transform.up);
                activePalm.relaxed();
            }
            else if(targetName.Contains("lshldr")) {
                // forward, up
                if(isRight)
                    handRotation = Quaternion.LookRotation(-target.transform.forward, target.transform.right);
                else
                    handRotation = Quaternion.LookRotation(-target.transform.forward, target.transform.right); 
                activePalm.grab();
            }
            else if(targetName.Contains("rshldr")) {
                // forward, up
                if(isRight)
                    handRotation = Quaternion.LookRotation(-target.transform.forward, -target.transform.right);
                else
                    handRotation = Quaternion.LookRotation(-target.transform.forward, -target.transform.right); 
                activePalm.grab();
            }
            else if(targetName.Contains("chest")) {
                // forward, up
                if(isRight) {
                    handRotation = Quaternion.LookRotation(target.transform.up, -directionVector);
                    //handRotation *= Quaternion.Euler(Vector3.up * 20);
                }
                else {
                    handRotation = Quaternion.LookRotation(target.transform.up, -directionVector); 
                    //handRotation *= Quaternion.Euler(Vector3.up * -20);
                }
                activePalm.squeezy();
            }
            else if(targetName.Contains("hip") || targetName.Contains("pelvis")) {
                // prawa ręka: forward jak - hip right,   up jak hip forward, 
                // lewa ręka: forward jak hip right, jak hip forward

                if(isRight)
                    handRotation = Quaternion.LookRotation(-target.transform.right, target.transform.forward);
                else
                    handRotation = Quaternion.LookRotation(target.transform.right, target.transform.forward);

                activePalm.relaxed();
            }
            else if(targetName.Contains("gen")) {
                // gotta know target sex!
                if(isRight)
                    handRotation = Quaternion.LookRotation(-target.transform.forward, -target.transform.right);
                else
                    handRotation = Quaternion.LookRotation(-target.transform.forward, target.transform.up);
                activePalm.grab();
            }
            else
                handRotation = Quaternion.LookRotation(target.up, target.right);
            
            // dłoń prawa - prosto = kciuk m w górę (BLUE Z), prawo = palce (RED X), góra = wierzch (GREEN Y). 
            // dłoń Lewa -  prosto                  (BLUE), prawo =-palce(-RED), góra = wierzch (GREEN). 

            // od nadgarstka do końca palca mam 20 cm
            // czyli dloń prawą cofamy w stronę nadgarstka: x-=0.15;
            // lewą x+=0.16

            // blue = z = prosto
            // red = x = prawo
            // green = y = góra

            activeHand.RBHoldPositionDamper = 100f;
            activeHand.RBHoldPositionSpring = 0f;

            var tighteningTime = 1.5f;
            var sustainTime = touchDuration - tighteningTime; 

            // hold position damper trzeba by na 100 dać...
            var pack = InterpolablePack.create()
            // rotate hand
            .add(new QuaternionInterpolator(activeHand.transform, handRotation, tighteningTime))
            // move hand
            .add(new VectorInterpolator(activeHand.transform, target.position, tighteningTime))
            // change controller tension
            // 0 - spręzyna nie działa, 1 - działa
            // czyli trzeba BARDZO powoli zwiększać do np 200 żeby dotknąć
            //                                 (current,                      target, inTime,           outTime, sustain, Action<float> l) : base (target, inTime, outTime, sustain) {
            .add(new LambdaEnvelopeInterpolator(activeHand.RBHoldPositionSpring, 1000, tighteningTime,   1f,    sustainTime, v => activeHand.RBHoldPositionSpring = v));

            bodyChannel.preempt(pack);
        }

        FreeControllerV3 prepareRightArmForReach() {
            return prepareArmForReach(rightHandController);
        }

        FreeControllerV3 prepareLeftArmForReach() {
            return prepareArmForReach(leftHandController);
        }

        FreeControllerV3 prepareArmForReach(FreeControllerV3 rightHandController) {
            rightHandController.currentPositionState = FreeControllerV3.PositionState.On;
            rightHandController.currentRotationState = FreeControllerV3.RotationState.On;

            // holdPositionSpring = 380 holdPositionDamper = 100  - sympatyczny, powolny ruch
            rightHandController.RBHoldPositionSpring = 800;
            rightHandController.RBHoldRotationDamper = 100;
            rightHandController.RBHoldRotationSpring = 50;
            // blisko celu, lub gdy przestanie się ruszać - zwiększać sprężynę aż do 2000

            // dłoń prawa - prosto = kciuk m w górę (BLUE Z), prawo = palce (RED X), góra = wierzch (GREEN Y). 
            // dłoń Lewa -  prosto                  (BLUE), prawo =-palce(-RED), góra = wierzch (GREEN). 

            // od nadgarstka do końca palca mam 20 cm
            // czyli dloń prawą cofamy w stronę nadgarstka: x-=0.15;
            // lewą x+=0.16

            // blue = z = prosto
            // red = x = prawo
            // green = y = góra

            // gładkie ruchy: https://docs.unity3d.com/ScriptReference/Vector3.SmoothDamp.html

            return rightHandController;
        }

        void stopReeachingWithArm(bool isRight) {
            if(isRight)
                controllerOff(rightHandController);
            else
                controllerOff(leftHandController);
        }


        void controllerOff(FreeControllerV3 rightHandController) {
            rightHandController.currentPositionState = FreeControllerV3.PositionState.Off;
            rightHandController.currentRotationState = FreeControllerV3.RotationState.Off;
        }




        string SaveFileCommentCharacter = "#";                                      //comment character in handpose files
        protected void LoadHandPose(Hand h, string path)
        {
            try
            {
                if (path.Length == 0) return;
                //read file
                string LoadedHandPose = SuperController.singleton.ReadFileIntoString(path);
                string[] LoadedHandPoseLines = LoadedHandPose.Split(new string[] { "\r\n", "\n" }, StringSplitOptions.None);
                foreach (string s in LoadedHandPoseLines)
                {
                    if (s.StartsWith(SaveFileCommentCharacter))
                    {
                        //its'a comment line, so ignore or send comment to logmessage
                        //if (Verbose) Log.d("HandPose: " + s);
                    }
                    else if (s.Length == 0)
                    {
                        //ignore empty lines
                    }
                    else
                    {
                        //split string into segment id and values array
                        string[] segValues = s.Split(new string[] { ":" }, StringSplitOptions.None);
                        string[] segmentID = segValues[0].Split(new string[] { " " }, StringSplitOptions.None);
                        string[] values = segValues[1].Split(new string[] { " " }, StringSplitOptions.None);
                        //use linq to find the segment in hand h
                        var finger = h.finger.Where(f => f.name == segmentID[0]).FirstOrDefault();
                        Segment segment;
                        if (finger == null)
                        {
                            segment = h.carpal.Where(c => c.name == segmentID[1]).FirstOrDefault();
                        }
                        else
                        {
                            segment = finger.segment.Where(seg => seg.name == segmentID[1]).FirstOrDefault();
                        }
                        if (segment == null)
                        {
                            //couldn't find a carpal or finger segment, do nothing
                        }
                        if (h.IsRight)
                        {
                            //set segment bone values to values in values array
                            segment.x = float.Parse(values[0]);
                            segment.y = float.Parse(values[1]);
                            segment.z = float.Parse(values[2]);
                        }
                        else
                        {
                            //set segment bone values to values in values array
                            //for left hand, invert Y and Z values
                            segment.x = float.Parse(values[0]);
                            segment.y = -float.Parse(values[1]);
                            segment.z = -float.Parse(values[2]);
                        }
                    }
                    h.UpdateRotations();
                }
            }
            catch (Exception e)
            {
                SuperController.LogError("LoadHandPose Exception " + e);
            }

        }
    }
}