using UnityEngine;
using UnityEngine.Events;
using UnityEngine.UI;
using UnityEngine.XR;
using System;
using System.Linq;

namespace Amaimon {

    class FingersCyclicInterpolator : SinusaidalInterpolator {
        Hand hand;

        public FingersCyclicInterpolator(Hand h, float perdiodSecs, float sustainPeriods=100000) : base(0.2f, perdiodSecs, sustainPeriods) {
            hand = h;
            _phase = 2;
        }

        override public void apply(float value) {
            // dostaniemy -0.2 do 0.2
            // 0.82 flat - 0.41 relaxed
            hand.openClose(value+0.41f);
        }
    }


    public partial class AActor : AAtom {

        SmartChannel smartHeadChannel;
        SmartChannel smartLeftHandChannel;
        SmartChannel smartRightHandChannel;
        SmartChannel smartPelvisChannel;

        AActor inLickRange = null;
        int headJobType = 0;

        void setupSmartPelvis() {
            smartPelvisChannel = new SmartChannel();
            smartPelvisChannel.noIdle();

            // if pelvis up is world up, then let's do balancing
            Func<float> idlePelvis = () => {
                //if(pelvisController.currentPositionState == FreeControllerV3.PositionState.Off) {

                    if(arousal >= 0.1f) {
                        var i = new HumpCycle(this, "pelvisControl", 10f, 2f/arousal, UnityEngine.Random.Range(1, 5));
                        smartPelvisChannel.run(i);
                        return i.getDuration();
                    }
                    
                    if(arousal < 0.8f) {
                        var period = UnityEngine.Random.Range(5, 10);
                        var i = new SwingCycle(this, "pelvisControl", 20f, period, 1);
                        smartPelvisChannel.run(i);
                        return i.getDuration();
                    }
                //}

                return UnityEngine.Random.Range(1f, 5f);
            };

            Func<float> idleChest = () => {
                var i = new HoolaCycle(this, "chestControl", 10f, 4f, UnityEngine.Random.Range(1f, 10f));
                smartPelvisChannel.run(i);
                return i.getDuration() + getScaredRelaxedInterval();
            };

            Func<float> lickRangeCheck = () => {

                inLickRange = null;

                StoryPlugin.ferEachActor(whom => {
                    if(whom != this && inLickRange == null) {
                        if(whom.genRb!=null) {
                            var adist = Vector3.Distance(whom.genitalPos.get(), headRb.position);  
                            if(adist <= 0.30f) {
                                inLickRange = whom;
                            }
                            //Log.d($"{person.name} Head pos={headRb.position}, {whom.person.name} gen pos={whom.genitalPos.get()} dist={adist}");
                        }
                    }          
                });

                if(inLickRange == null && headJobType != 0) {
                    Log.d($"{person.name} Out of lick range");
                    smartHeadChannel.setMainJob(mainHeadJob);
                    headJobType = 0;
                }
                else if(inLickRange != null && headJobType != 1) {
                    Log.d($"{person.name} In lick range");
                    expressionChannel.stop();
                    smartHeadChannel.setMainJob(lickHeadJob);
                    headJobType = 1;
                }

                return 1.5f;
            };

            smartPelvisChannel.addIdleJob(idlePelvis);
            if(!isMale())
                smartPelvisChannel.addIdleJob(idleChest);
            smartPelvisChannel.addIdleJob(lickRangeCheck);

        }
        void setupSmartHands() {
            smartLeftHandChannel = new SmartChannel();
            smartLeftHandChannel.noIdle();

            smartRightHandChannel = new SmartChannel();
            smartRightHandChannel.noIdle();

            Func<float> rHand = () => {
                if(manageTouch.val)
                {
                    controllerOff(rightHandController);

                    var movementDuration = UnityEngine.Random.Range(1f, 5f);
                    var withPause = movementDuration + UnityEngine.Random.Range(1f, 5f);

                    var i = new ForceCyclicInterpolator(this, "rHandControl", Vector3.forward, 0.5f, UnityEngine.Random.Range(1f, 5f));
                    smartLeftHandChannel.run(i);
                    return withPause;
                }
                return 2;
            };

            Func<float> lHand = () => {
                if(manageTouch.val) {
                    controllerOff(leftHandController);

                    var movementDuration = UnityEngine.Random.Range(1f, 5f);
                    var withPause = movementDuration + UnityEngine.Random.Range(1f, 5f);

                    var i = new ForceCyclicInterpolator(this, "lHandControl", Vector3.forward, 0.5f, UnityEngine.Random.Range(1f, 5f));
                    smartLeftHandChannel.run(i);
                    return withPause;
                }
                return 2;
            };

            Func<float> rFingers = () => {
                var movementDuration = UnityEngine.Random.Range(1f, 10f);
                var withPause = movementDuration + UnityEngine.Random.Range(1f, 5f);

                var i = new FingersCyclicInterpolator(rightHand,3,UnityEngine.Random.Range(1f, 3f));
                smartLeftHandChannel.run(i);

                return withPause;
            };


            Func<float> lFingers = () => {
                var movementDuration = UnityEngine.Random.Range(1f, 10f);
                var withPause = movementDuration + UnityEngine.Random.Range(1f, 5f);

                var i = new FingersCyclicInterpolator(leftHand,3,UnityEngine.Random.Range(1f, 3f));
                smartLeftHandChannel.run(i);

                return withPause;
            };

            Func<float> lInterest = () => {
                var movementDuration = 0f;
                try {
                    if(manageTouch.val)
                        movementDuration=prioritizeHands(false);
                } catch (Exception ex) {
                    Log.d($"Błąd w lInterest pio {person.name}:{ex}");
                }

                return movementDuration+getScaredRelaxedInterval();
            };


            Func<float> rInterest = () => {
                var movementDuration = 0f;
                try {
                    if(manageTouch.val)
                        movementDuration=prioritizeHands(true);
                } catch (Exception ex) {
                    Log.d($"Błąd w rInterest pio {person.name}:{ex}");
                }

                return movementDuration+getScaredRelaxedInterval();
            };

            smartLeftHandChannel.addIdleJob(lHand);
            smartLeftHandChannel.addIdleJob(lFingers);
            smartLeftHandChannel.setMainJob(lInterest);

            smartRightHandChannel.addIdleJob(rHand);
            smartRightHandChannel.addIdleJob(rFingers);
            smartRightHandChannel.setMainJob(rInterest);
        }



        Func<float> lickHeadJob;

        Func<float> mainHeadJob;

        void setupSmartHead() {
            smartHeadChannel = new SmartChannel();
            smartHeadChannel.noIdle();

            lickHeadJob = () => {
//Log.d($"{person.name} Running main lick job");
                faceTowards(inLickRange.genRb.position, true, 2);
                eyesTowards(inLickRange.headRb.position, true);

                if(lickMorph == null) {
                    Log.d("Creating lick morph");
                    lickMorph = getLickingMorph();
                }
                smartHeadChannel.overrideMainInterpolable(lickMorph, true, 10000);
                return 10000;
            };

            mainHeadJob = () => {
                var pause = 0f;
//Log.d($"{person.name} Running main head job");
                try {
                    if(manageInterest.val)
                        pause = prioritizeInterestPoints();
                } catch (Exception ex) {
                    Log.d($"Błąd w headInterest pio {person.name}:{ex}");
                }

                return pause + getScaredRelaxedInterval();
            };


            Func<float> saccade = () => {
                if(interestDoSaccade) {
                    var dist = Vector3.Distance(headRb.transform.position, interestLocation);
                    var range = dist / 25f;
                    var yshift = Vector3.up * UnityEngine.Random.Range(-range, range);
                    var xshift = Vector3.right * UnityEngine.Random.Range(-range, range);

                    eyeTarget.transform.position = interestLocation + yshift + xshift;
                }

                checkPreempt();

                return 0.2f;
            };

            Func<float> headMicro = () => {
                if(manageHead.val) {
                    setJointStateForXatarina(headControl, FreeControllerV3.PositionState.Comply, FreeControllerV3.RotationState.Comply);
                    headControl.RBHoldRotationSpring = 15;
                    headControl.RBHoldRotationDamper = 8; 

                    var noise = Quaternion.Euler(UnityEngine.Random.Range(-2f,2f),UnityEngine.Random.Range(-2f,2f),UnityEngine.Random.Range(-5f,5f));
                    var lookDirection = headControl.transform.rotation * noise ; // up żeby był 90 stopni do kierunku

                    smartHeadChannel.run(new QuaternionInterpolator(headControl.transform, lookDirection, 1f));
                }
                return 1f;
            };

            Func<float> moodExpression = () => {
                if(manageEmotions.val) {
                    return chooseExpression() + UnityEngine.Random.Range(0f, getScaredRelaxedInterval()*2);
                }
                else
                {
                    return 5;
                }
            };

            smartHeadChannel.addIdleJob(saccade);
            smartHeadChannel.addIdleJob(headMicro);
            smartHeadChannel.addIdleJob(moodExpression);
            smartHeadChannel.setMainJob(mainHeadJob);
        }

        float blinkProgress = 0f;
        float untilNextBlink = 5f;
        float blinkDuration = 0.15f;
        float currentEyelidPower = 0f;

        void eyelid(float deltaT) {
            var angle = eyeBone.transform.localRotation.eulerAngles;
            // mgnienie oka = 100 ms, zastosujmy abs(cosinus(0 : pi))
            // odstępy = 20 mgnięć na minutę = co 3 sekundy
            // czyli w 100 ms musimy przejść od 0 do pi

            // 100ms == pi
            // deltaT == x

            untilNextBlink -= deltaT;

            if(untilNextBlink <= 0) {
                // blink in progress, process only blink
                blinkProgress += deltaT;

                float blinkPhase = (blinkProgress * Mathf.PI) / blinkDuration;
                float blinkCoef = currentEyelidPower + Mathf.Abs(Mathf.Sin(blinkPhase)); // 0 - open, 1 - closed

                eyelidMorph.morphValue = Mathf.Clamp(blinkCoef, 0f, 1f);
            }
            else {
                // not blinking at the moment, process eyelid

                // kat 15 - 0.150 eyes closed REN
                //     27 - 0.500
                // do daje: (0.5 - 0.15) / (27 - 15) = 0,35 / 12 = 0,29....
                var corrAngle = angle.x;
                if(corrAngle >= 180)
                    corrAngle = corrAngle - 360f;
                
                currentEyelidPower = Mathf.Clamp(corrAngle * 0.29f -4.2f, 0f, 0.6f);            

                eyelidMorph.morphValue = currentEyelidPower;
            }

            if(blinkProgress >=blinkDuration) {
                // end of eye blink, begin pause
                blinkProgress = 0;
                untilNextBlink = UnityEngine.Random.Range(2.5f, 6f);
            }



        }

        public LifecycleInterpolable getLickingMorph() {
            // AA - Tongue 1 = 0.64 (0.67 do dt)
            // Tongue Bend Tip = -0.686 - 0.187 albo tongue XXX 0.150-0.464 (default = 0)

            // OnOffMorph(AActor person, string name, float onVal, float offVal, float inTime, float outTime) : base(onVal, offVal, inTime, outTime) {

            Func <float,float> bendFn =  (x) =>  {
                 // return perlin noise in range -0.686 - 0.187 ==> zakers 0.873
                float rn = Mathf.PerlinNoise(Time.time * 2, 0) * 0.873f;
                return rn - 0.686f;
            } ;

            return ConcurrentInterpolablePack.create()
                .add(new OnOffMorph(this,"AA - Tongue 1", 0.65f, 0f, 1.5f, 1f))
                .add(new OnOffMorph(this,"Tongue Bend Tip", 0.187f, 0f, 2f, 0.5f, bendFn))
                ;
        }

        RandomTable<LifecycleInterpolable> getGenitalCaressMorphs() {
            return RandomTable<LifecycleInterpolable>.create()
                .add(getLickingMorph(), 1f)
                .add(new ExpressionMorph(this, "AA - Kiss 1", 1f, 1f, 0.5f), 1f)
                .add(new ExpressionMorph(this, "AA - Kiss 2", 1f, 1f, 0.5f), 1f)
                .add(new ExpressionMorph(this, "AA - Kiss 3", 1f, 1f, 0.5f), 1f)
                ;
        }

        LifecycleInterpolable lickMorph = null;
    }
}