using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;
using UnityEngine;
using UnityEngine.Events;
using UnityEngine.UI;
using UnityEngine.XR;
using UnityEngine.Windows.Speech;

using SimpleJSON;

namespace Amaimon {
    public class AAtom  {
        public Atom person;
        protected Hitter lastHit = null;
        protected Hitter lastITouched = null;

        ///////////////////////////////////////////////////
        // Konstruktory
        ///////////////////////////////////////////////////

        public AAtom() {
        }

        public AAtom(Atom a) {
            if(a == null)
                throw new NullReferenceException("Null storables for AArom constructor"); 
            person = a;
        }

        public AAtom(string name, MVRScript context) {
            person = context.GetAtomById(name);
            if(person == null)
                SuperController.LogError("Couldn't find atom ${name}!");
        }



        ///////////////////////////////////////////////////
        // Propertasty
        ///////////////////////////////////////////////////

        public Vector3 position {
            get {
                return getController("control").transform.position;
            }
            set {
                getController("control").transform.position = value;
                //person.ResetPhysics(true, true);
            }
        }

        public Quaternion rotation {
            get {
                return getController("control").transform.rotation;
            }
            set {
                getController("control").transform.rotation = value;
            }
        }



        ///////////////////////////////////////////////////
        // Kwerendy
        ///////////////////////////////////////////////////

        public virtual float getDistance(string ofWhat, AAtom otherActor, string otherPart) {
            Vector3 pers2node = otherActor.fuzzyGetTransform(otherPart).position;
            Vector3 pers1node = fuzzyGetTransform(ofWhat).position;
            return Vector3.Distance(pers1node, pers2node);            
        }

        public virtual bool isLookingTowards(AAtom whom, string what, float fov, float diff = 1000f) {
            return false;
        }

        public Rigidbody getRigidBody(string name) {
            return person.rigidbodies.FirstOrDefault(fc => fc.name == name);
        }

        public FreeControllerV3 getController(string name) {
            return person.freeControllers.FirstOrDefault(fc => fc.name == name);
        }

        public virtual UniversalRigidbody getUniversalRigidbody(string name) {
            return new RealRigidbody(getRigidBody(name), true);
        }

        public virtual Transform fuzzyGetTransform(string nodeName){
            Transform nodePosition = null;	

            foreach (FreeControllerV3 fc in person.freeControllers){ 
                if (fc.name=="control" && nodeName=="control"){nodePosition=fc.transform;}
                else if (fc.name==nodeName+"Control"){nodePosition=fc.followWhenOff;}
                else if (fc.name==nodeName){nodePosition=fc.transform;}
            }

            foreach (Rigidbody rb in person.rigidbodies){
                if (rb.name==nodeName && nodeName.Contains("Trigger")){nodePosition=rb.transform;}
                else if(rb.name==nodeName){nodePosition=rb.transform;}
            }

            return nodePosition;
        }

        public T GetComponentByName<T>(string name) where T: Component{
            var components = person.GetComponentsInChildren<T>();

            foreach(T comp in components) {
                if(comp.name == name)
                    return comp;
            }            
            return null;   
        }


        public virtual Component[] GetComponentsInChildren(Type t) {
            return person.GetComponentsInChildren(t);   
        }

        public virtual string getName() {
            return person.name;
        }



        ///////////////////////////////////////////////////
        // Collidery
        ///////////////////////////////////////////////////
        virtual public void unregisterColliders() {
            Log.d($"unregistering CollisionAnalyzers for {person.name}");
            foreach(var component in person.GetComponentsInChildren<CollisionAnalyzer>()) {
                Log.d($"destroing {person.name}'s {component}");
                component.flush();
                UnityEngine.Object.Destroy(component);
            }
        }

        virtual public void registerColliders() {
            unregisterColliders();

            getHitters().ForEach(hitter => {
                    attachCollisionCallback(hitter.gameObject, imCollidingWith);
                });
        
        }

        void attachCollisionCallback(GameObject offenderPart, EventHandler<CollisionEventArgs> callHere) {
            var collider = offenderPart.AddComponent<CollisionAnalyzer>();
            collider.OnCollide += callHere;
            collider.offender = this;
            collider.offendersPart = offenderPart;
        }

        virtual public void imCollidingWith(object analyzer, CollisionEventArgs e)
        {
            var whomDidIHit = detectVictim(e);
            if(whomDidIHit != this) {
                var hitter = new Hitter(e.offender, e.offendersPart.name.ToLower());
                hitter.victimBodyPart = e.collision.rigidbody.niceName();
                hitter.hitType = e.evtType;

                whomDidIHit?.wasHitWith(hitter);
            }
        }

        virtual public void wasHitWith(Hitter h) {
            Log.d("wasHitWith called in aatom");
        }


        public AActor detectVictim(CollisionEventArgs e) {
            foreach(AAtom interactable in StoryPlugin.storyObjects) {
                foreach (Collider rb in interactable.GetComponentsInChildren(typeof(Collider))){
                    if(rb == e.collision.collider && interactable is AActor) return (AActor)interactable;
                }   
            }
            return null;
        }



        ///////////////////////////////////////////////////
        // Interakcja
        ///////////////////////////////////////////////////

        // list parts that might be interesting for the actors
        virtual public List<UniversalRigidbody> getInterests() {
            var list = new List<UniversalRigidbody>();

            try {
                var rb = new RealRigidbody(person.rigidbodies[0], true);
                list.Add(rb);
            } catch (Exception ex) {

            }
            return list;
        }

        // list parts that actors may touch
        virtual public List<UniversalRigidbody> getTouchTargets() {
            return getInterests();
        }

        // List parts that might be used to hit other actors
        virtual public List<UniversalRigidbody> getHitters() {
            var list = new List<UniversalRigidbody>();

            if(person.category == "Toys" || person.category == "Shapes" || person.name.StartsWith("IS")) {
                if (person.rigidbodies.Length > 0) { 
                    foreach (Rigidbody rigidbody in person.rigidbodies) {
                        list.Add(new RealRigidbody(rigidbody)); 
                    }
                }
            }
            else if(person.type == "CustomUnityAsset") {
                foreach (Transform transform in person.GetComponentsInChildren<Transform>())
                {
                    // if (transform.GetComponent<Collider>() || transform.GetComponent<MeshCollider>() || transform.GetComponent<CapsuleCollider>() || transform.GetComponent<BoxCollider>() || transform.GetComponent<NonConvexMeshCollider>() || transform.GetComponent<SphereCollider>())
                    // {
                    //     list.Add(transform.gameObject);
                    // }
                }
                if (person.rigidbodies.Length > 0)
                {
                    foreach (Rigidbody rigidbody in person.rigidbodies)
                    {
                        list.Add(new RealRigidbody(rigidbody));
                    }
                }
            }

            return list;
        }

        virtual public void cleanup() {
            Log.d($"cleanup w obiekcie:{person.name}");
            unregisterColliders();
        }

        ///////////////////////////////////////////////////
        // Atom utilities
        ///////////////////////////////////////////////////

        public void parentLinkAtPosition(Atom parent, JSONStorable toPosition, string toPositionName)
        {
            person.parentLinkAtPosition(parent, toPosition, toPositionName);
        }

        protected virtual IEnumerator CreateAtom(string atomType, string atomId, Action<Atom> onAtomCreated)
        {
            Atom atom = SuperController.singleton.GetAtomByUid(atomId);
            if (atom == null)
            {
                yield return SuperController.singleton.AddAtomByType(atomType, atomId);
                atom = SuperController.singleton.GetAtomByUid(atomId);
            }
            if (atom != null)
            {
                onAtomCreated(atom);
            }
        }

        public void delete() {
            UnityEngine.Object.Destroy(person);
        }

        public void setCollision(bool state) {
            person.collisionEnabledJSON.val = state;
        }

        public T registerPlugin<T> () where T : APlugin  {
            Type type = typeof(T);
            T created = (T)Activator.CreateInstance(type, this);
            if(created.storables != null) {
                created.init();
                return created;
            }
            else {
                //Log.d($"WARNING!!! {person.name} could not register ${type}");
                return null;
            }
        }



















        public void dumpStorables() {
            person?.GetStorableIDs()?.ForEach(a => Log.d($"Storable: {a}"));
        }

        public string sanitizeName(string trimmedPluginName) {
            return Regex.Replace(trimmedPluginName, "[^A-Za-z0-9]", "");
        }

        public string propertyToC(string type) {
            if(type=="BoolProperty")
                return "bool";
            else if(type=="StringProperty" || type=="StringChooserProperty")
                return "string";
            else if(type=="FloatProperty")
                return "float";
            else if(type=="ActionProperty")
                return "action";
            else if(type=="Url")
                return "Url";
            else if(type=="FilePathAction")
                return "action";
            else if(type=="ColorProperty")
                return "Color";
            else  {
                SuperController.LogError($"Don't know how to process to C {type}");
                throw new NotImplementedException();
            }
        }

        public string jsonToNamedProperty(JSONStorable.Type type) {
            /*
            		None,
		Vector3,
		Color,
		AudioClipAction,
		SceneFilePathAction,
		StringChooserAction
        */
            if(type==JSONStorable.Type.Bool)
                return "BoolProperty";
            else if(type==JSONStorable.Type.String)
                return "StringProperty";
            else if(type==JSONStorable.Type.StringChooser)
                return "StringChooserProperty";
            else if(type==JSONStorable.Type.Float)
                return "FloatProperty";
            else if(type==JSONStorable.Type.Action)
                return "ActionProperty";
            else if(type==JSONStorable.Type.Url)
                return "Url";
            else if(type==JSONStorable.Type.PresetFilePathAction)
                return "ActionProperty";
            else if(type==JSONStorable.Type.Color)
                return "ColorProperty";
            else {
                SuperController.LogError($"Don't know how to process storable {type}");
                throw new NotImplementedException();
            }
        }

        public void preparePluginAPI() {
            List<string> fieldList = new List<string>();
            List<string> cleanFields = new List<string>();
            List<string> NamedProperties = new List<string>();

            List<string> FileLines = new List<string>();

            Transform pluginsTransform = person.transform.Find("reParentObject")?.Find("object")?.Find("PluginManager")?.Find("Plugins");

            for (int i = 0; i < pluginsTransform.childCount; i++){
                string receiverName = pluginsTransform.GetChild(i).gameObject.name;
                JSONStorable receiver = person.GetStorableByID(receiverName);
                int first = receiverName.IndexOf("_");
                string trimmedPluginName; 
                if(first>0)
                    trimmedPluginName = receiverName.Substring(first+1);
                else
                    trimmedPluginName = receiverName;

                string cleanPluginName = sanitizeName(trimmedPluginName);

                FileLines.Add($"public class {cleanPluginName}Plugin : APlugin {{");
                FileLines.Add($"    public {cleanPluginName}Plugin(AActor actor) : base(actor, \"{trimmedPluginName}\") {{ }}");

                fieldList.Clear();
                cleanFields.Clear();
                NamedProperties.Clear();

                if(receiver != null) {
                    List<string> lista = receiver.GetAllParamAndActionNames();
                    if(lista.Count > 17) {
                        for(int j = 17; j<lista.Count; j++) {

                            string name = lista[j];
                            JSONStorable.Type type = receiver.GetParamOrActionType(name);
                            fieldList.Add(name);
                            cleanFields.Add(sanitizeName(name));
                            NamedProperties.Add(jsonToNamedProperty(type));

                        }
                    }

                    if(fieldList.Count > 0) {
                        FileLines.Add("    override public void init() {");

                        for(int j = 0; j<fieldList.Count; j++) {
                            FileLines.Add($"        {cleanFields[j]} = new {NamedProperties[j]}(this,\"{fieldList[j]}\");");
                        }

                        FileLines.Add("    }");

                        for(int j = 0; j<fieldList.Count; j++) {
                            string cType = propertyToC(NamedProperties[j]);

                            if(cType == "action") {

                            string oneEntry = @"	
    public "+NamedProperties[j]+@" "+cleanFields[j]+@";
";
                            FileLines.Add(oneEntry);
                            }
                            else {
                            string oneEntry = @"	
    private "+NamedProperties[j]+@" "+cleanFields[j]+@";

	public void set"+cleanFields[j]+@"("+cType+@" a) {
		"+cleanFields[j]+@".set(a);
	}
	
	public "+cType+@" get"+cleanFields[j]+@"() {
		return "+cleanFields[j]+@".get();
	}
";
                            FileLines.Add(oneEntry);

                            }
                        }
                    }
                }

                FileLines.Add("}");
            }
            
            FileWriter.write(FileLines, _ini);
        }

        private string _ini = "\\modules.cs";


 
    }
}