using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;

using UnityEngine;
using UnityEngine.Events;
using UnityEngine.UI;
using UnityEngine.XR;
using UnityEngine.Windows.Speech;

using SimpleJSON;

using MVR.FileManagementSecure;

namespace Amaimon {
    public class Player : AAtom {
        private static Player _instance = null;
        private KeywordRecognizer _recognizer = null;
        private DictationRecognizer _dRecognizer = null;
        List<string> phraseDictionary = new List<string>();

        public GameObject playerChest;
        GameObject playerBelly;
        GameObject playerGen;
        public GameObject playerHead;
        GameObject playerHands;

        string[] fingers = new string[] 
        {
            "Thumb", "Index", "Middle", "Ring", "Pinky"
        };

        bool _leftStateResetButtonPressed = false;

        bool _leftStateChangeButtonPressed = false;
        int _leftHandStateIndex = -1;

        bool _rightStateResetButtonPressed = false;

        bool _rightStateChangeButtonPressed = false;
        int _rightHandStateIndex = -1;

        Dictionary<int, string> handStates;
        List<UniversalRigidbody> rigidbodies;
        private UnityEngine.Windows.Speech.DictationRecognizer.DictationResultDelegate dictationResultDelegate;
        private Action<string> currentSpeechCallback = null;


        private Player() : base() {}

        public void setup() {   
            // setup flags
            SuperController.singleton.commonHandModelControl.useCollision = true;
            SuperController.singleton.allowGrabPlusTriggerHandToggle = false;

            // setup player body
            StoryPlugin.singleton.StartCoroutine(addFakeBodyParts());

            // setup hand states
            try
            {
                handStates = new Dictionary<int, string>();

                handStates.Add(0, "Index");
                handStates.Add(1, "Middle");
                handStates.Add(2, "Ring");
                handStates.Add(3, "Pinky");
                handStates.Add(4, "Thumb");
                handStates.Add(5, "ThumbIndex");
                handStates.Add(6, "IndexMiddle");
                handStates.Add(7, "IndexMiddleRing");
                handStates.Add(8, "IndexMiddlePinky");
                handStates.Add(9, "IndexRingPinky");
                handStates.Add(10, "Fist");
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception while adding hands: " + e);
            }

            startSentenceRecognizer();
        }

        public void removeFakeBodyParts() {
            unregisterColliders();

            if(playerChest)
                UnityEngine.Object.Destroy(playerChest);
            
            if(playerBelly)
                UnityEngine.Object.Destroy(playerBelly);
            
            if(playerGen)
                UnityEngine.Object.Destroy(playerGen);

            if(playerHead)
                UnityEngine.Object.Destroy(playerHead);

        }

        public void celanup() {
            stopSentenceRecognizer();

            removeFakeBodyParts();

            _instance = null;
        }

        public static Player singleton
        {
            get
            {
                if (_instance == null){_instance = new Player();}
                return _instance;
            }
        }


        // Update is called with each rendered frame by Unity
        public void Update()
        {
            if(!playerHands)
                return;
            try
            {
                if(SuperController.singleton.GetLeftSelect())
                    HandleUserInputStateChangeButton(ref _leftStateChangeButtonPressed, ref _leftHandStateIndex);

                if (SuperController.singleton.GetGrabNavigateLeft())
                    HandleUserInputStateResetButton(ref _leftStateResetButtonPressed, ref _leftHandStateIndex);

                if (SuperController.singleton.GetRightSelect())
                    HandleUserInputStateChangeButton(ref _rightStateChangeButtonPressed, ref _rightHandStateIndex);

                if (SuperController.singleton.GetGrabNavigateRight())
                    HandleUserInputStateResetButton(ref _rightStateResetButtonPressed, ref _rightHandStateIndex);
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        // FixedUpdate is called with each physics simulation frame by Unity
        public void FixedUpdate()
        {
            if(!playerHands)
                return;
            try
            {
                if (_leftStateChangeButtonPressed)
                    _leftStateChangeButtonPressed = false;

                if(_leftStateResetButtonPressed)
                    _leftStateResetButtonPressed = false;
                
                if (_rightStateChangeButtonPressed)
                    _rightStateChangeButtonPressed = false;

                if (_rightStateResetButtonPressed)
                    _rightStateResetButtonPressed = false;

                ChangeHandState("L", _leftHandStateIndex);
                ChangeHandState("R", _rightHandStateIndex);
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }


       public IEnumerator addFakeBodyParts() {
           Log.d("Add fake rb started");
            rigidbodies = new List<UniversalRigidbody>();

            playerHands = SuperController.singleton.commonHandModelControl.gameObject;
            //var playerHands = GameObject.Find("Hands");
            if (playerHands)
            {
                Rigidbody lhand = null;
                Rigidbody rhand = null;

                playerHands
                    .GetComponentsInChildren<Rigidbody>().ForEach(rb => {

                if(rb.name == "HandPhysical" && lhand == null) {
                    //Log.d($"Core control rigid left hand:{rb.name}");
                    lhand = rb;
                    rigidbodies.Add(new RenamedRealRigidbody("lHand", lhand));
                }
                else if(rb.name == "HandPhysical" && rhand == null) {
                    //Log.d($"Core control rigid right hand:{rb.name}");
                    rhand = rb;
                    rigidbodies.Add(new RenamedRealRigidbody("rHand", rhand));
                }
                else if(rb.name != "HandPhysical")
                    rigidbodies.Add(new RealRigidbody(rb));

                });
            }

            playerHead = createGemeObjectWithCollider("playerHead", 0.00f, 0.25f, 0.08f);
            playerChest = createGemeObjectWithCollider("playerChest", 0.30f, 0.35f, 0.15f);
            playerBelly = createGemeObjectWithCollider("playerBelly", 0.60f, 0.20f, 0.15f);
            playerGen = createGemeObjectWithCollider("playerGen3", 0.85f, 0.10f, 0.02f);

            yield return new WaitForSeconds(5);

            if(playerHead != null)
                rigidbodies.Add(new RenamedRealRigidbody("head", playerHead.GetComponent(typeof(Rigidbody)) as Rigidbody, true));
            else
                Log.d("Coudln't create head");

            if(playerChest != null)
                rigidbodies.Add(new RenamedRealRigidbody("chest", playerChest.GetComponent(typeof(Rigidbody)) as Rigidbody, true));
            else
                Log.d("Coudln't create chest");

            if(playerBelly != null)
                rigidbodies.Add(new RenamedRealRigidbody("abdomen", playerBelly.GetComponent(typeof(Rigidbody)) as Rigidbody, true));
            else
                Log.d("Coudln't create belly");

            if(playerGen != null)
                rigidbodies.Add(new RenamedRealRigidbody("gen3", playerGen.GetComponent(typeof(Rigidbody)) as Rigidbody, true));
            else
                Log.d("Coudln't create uj");



            // Log.d($"lhand={lhand}, rhand={rhand}, all player bodies={rigidbodies.Count()}");

            registerColliders();

            StoryPlugin.ferEachActor(act => act.applySettingsFromUI());
           Log.d("Add fake rb finished");
       }


        override public UniversalRigidbody getUniversalRigidbody(string name) {
            try {
                return rigidbodies.Where(rb => rb.name == name).First();
            } catch (Exception ex) {
                //Log.d($"Player couldn't find {name} rb");
                return null;
            }
        }


        // Transform getLeftHandTransform() {return(SuperController.singleton.leftHand);}
        // Transform getRightHandTransform() {return(SuperController.singleton.rightHand);}
        public Transform getHeadTransform() {
            //var test = new Transform();
            // test.position = new Vector3(10,20,30);

            // test.SetParent(CameraTarget.centerTarget.transform);


            return(CameraTarget.centerTarget.transform) ;
        } // also have Transform .tansfrom proprty

        GameObject createGemeObjectWithCollider(string name, float headOffset, float height, float radius) {
            var newBody = GameObject.Find(name);
            if(newBody != null)
                UnityEngine.Object.Destroy(newBody);

            newBody = new GameObject(name);

            //var newBody = GameObject.Find("[CameraRig]");

            Rigidbody rb = newBody.AddComponent<Rigidbody>();
            rb.isKinematic = true;

            var headPos = CameraTarget.centerTarget.transform.position;
            newBody.transform.position = headPos - Vector3.up * headOffset; // 30 - piersi, 60 brzuch, 85 - chuj
            //Log.d($"{name} position: {newBody.transform.position}");
            newBody.transform.SetParent(CameraTarget.centerTarget.transform);

            // CapsuleCollider bc = newBody.AddComponent<CapsuleCollider>();
            // bc.center = Vector3.zero;
            // bc.height = height;
            // bc.radius = radius;
            // bc.direction = 1;

            return newBody;
        }

        override public Transform fuzzyGetTransform(string nodeName){
            try {
                return rigidbodies.Where(rb => rb.niceName().Contains(nodeName)).First().transform;
            } catch (Exception ex) {
                return null;
            }
        }

        override public List<UniversalRigidbody> getInterests() {
            var list = new List<UniversalRigidbody>();

            //Log.d("Getting what's interesting in the Player...");

            UniversalRigidbody hand = null;
            hand = getUniversalRigidbody("head");

            if(hand != null)
                list.Add(hand);

            hand = getUniversalRigidbody("chest");

            if(hand != null && hand.transform != null) {
                list.Add(hand);
            }

            hand = getUniversalRigidbody("abdomen");

            if(hand != null && hand.transform != null) {
                list.Add(hand);
            }

            hand = getUniversalRigidbody("gen3");

            if(hand != null && hand.transform != null) {
                list.Add(hand);
            }

            hand = getUniversalRigidbody("lHand");

            if(hand != null)
                list.Add(hand);
            
            hand = getUniversalRigidbody("rHand");

            if(hand != null)
                list.Add(hand);

            //Log.d($"Interests in player:{list.Count()}");

            //list.ForEach(i => Log.d($"player intetrest:{i?.name} <<<"));

            return list;
        }

        override public List<UniversalRigidbody> getTouchTargets() {
            return getInterests();
        }

        override public List<UniversalRigidbody> getHitters() {
            var list = new List<UniversalRigidbody>();

            //Log.d("Getting what Player can hit with");

            UniversalRigidbody hand = null;

            hand = getUniversalRigidbody("head");

            if(hand != null)
                list.Add(hand);

            hand = getUniversalRigidbody("lHand");

            if(hand != null)
                list.Add(hand);
            
            hand = getUniversalRigidbody("rHand");

            if(hand != null)
                list.Add(hand);

            list.AddRange(rigidbodies.Where(rb => rb.name.Contains("_Finger_")));

            return list;
        }

        override public Component[] GetComponentsInChildren(Type t) {
            var hands = SuperController.singleton.commonHandModelControl.gameObject;
            if(hands != null) {
                return hands.GetComponentsInChildren(t);
            }
            else
                return null;
        }

       override public string getName() {
            return "CoreControl";
        }

        override public void unregisterColliders() {
            var hands = SuperController.singleton.commonHandModelControl.gameObject;

// transform rąk:HeightOffset (UnityEngine.GameObject)...
// transform rąk:[CameraRig] (UnityEngine.GameObject)... - podczepimy się pod camera rig, zamiast tworzyć nowe gameobjecty?
// transform rąk:NavRig (UnityEngine.GameObject)...
// transform rąk:WorldScaleAdjust (UnityEngine.GameObject)...
// transform rąk:CoreControl (UnityEngine.GameObject)...
// transform rąk:SceneAtoms (UnityEngine.GameObject)...

            // var tr = hands.GetComponent<Transform>().parent.gameObject;
            // do {
            //     Log.d($"transform rąk:{tr}...");
            //     tr = tr.GetComponent<Transform>().parent.gameObject;
            // } while(tr != null);

            if(hands != null) {
                foreach(var component in hands.GetComponentsInChildren<CollisionAnalyzer>()) {
                    //Log.d($"destroing {person.name}'s {component}");
                    component.flush();
                    UnityEngine.Object.Destroy(component);
                }
            }
        }

        override public void registerColliders() {
            unregisterColliders();

            getHitters().ForEach(hitter => {
                //Log.d($"Pro:{hitter}");
                attachCollisionCallback(hitter.gameObject, imCollidingWith);
            });
        
        }

        void attachCollisionCallback(GameObject offenderPart, EventHandler<CollisionEventArgs> callHere) {
            var collider = offenderPart.AddComponent<CollisionAnalyzer>();
            collider.OnCollide += callHere;
            collider.offender = this;
            collider.offendersPart = offenderPart;
        }

        void imCollidingWith(object analyzer, CollisionEventArgs e)
        {
            var whomDidIHit = detectVictim(e);
            var hitter = new Hitter(e.offender, e.offendersPart.name.ToLower());
            hitter.victimBodyPart = e.collision.rigidbody.niceName();
            hitter.hitType = e.evtType;

            whomDidIHit?.wasHitWith(hitter);
        }

        AActor detectVictim(CollisionEventArgs e) {
            foreach(AAtom interactable in StoryPlugin.storyObjects) {
                foreach (Collider rb in interactable.GetComponentsInChildren(typeof(Collider))){
                    if(rb == e.collision.collider && interactable is AActor) return (AActor)interactable;
                }   
            }
            return null;
        }

        override public float getDistance(string ofWhat, AAtom otherActor, string otherPart) {
            Vector3 pers1node = fuzzyGetTransform(ofWhat).position;

            if(pers1node == null)
                return 10000;

            Vector3 pers2node = otherActor.fuzzyGetTransform(otherPart).position;
            return Vector3.Distance(pers1node, pers2node);            
        }

        // diff can be used to check directionality. I.e. if you're checking pelvis
        // setting diff to 90 will return true if the look comes from behind - I mean ASS
        override public bool isLookingTowards(AAtom whom, string what, float fov, float diff = 1000f) {
            Transform targetTransform = whom.fuzzyGetTransform(what);
            float headToTargetAngle = QuatOper.controlToSomethingAngle(CameraTarget.centerTarget.transform, targetTransform.position);

            float directionDiff = Vector3.Angle(targetTransform.forward, CameraTarget.centerTarget.transform.forward);

            return (headToTargetAngle<=(fov/2f)) && (directionDiff < diff);
        }


        // left/right hand są tylko Transform...

        public long GetTimeNow(bool scaledTime = false){long msTime ; if (!scaledTime) { msTime = (long)(Time.unscaledTime * 1000); } else { msTime = (long) (Time.time *1000); }return (msTime);}

        public void doOnVoiceAnswer(List<string> phrases, Func<string, bool> action) {
            isListeningForDictionary = false;
            phraseDictionary = phrases;
            _recognizer = new KeywordRecognizer (phraseDictionary.ToArray());_recognizer.OnPhraseRecognized += (args) => {
                action(args.text);
            };
            _recognizer.Start();
        }

        public bool isListeningForDictionary {
            set {
                if (_recognizer != null){ _recognizer.OnPhraseRecognized -= OnPhraseRecognized;_recognizer.Stop();PhraseRecognitionSystem.Shutdown();_recognizer = null; }

                if(value) {
                    _recognizer = new KeywordRecognizer (phraseDictionary.ToArray());_recognizer.OnPhraseRecognized += OnPhraseRecognized;_recognizer.Start();
                }
            }
        }





        public void stopSentenceRecognizer() {
            /*_dRecognizer.OnPhraseRecognized -= DictationResult;*/
            if(_dRecognizer != null) {
                _dRecognizer.DictationResult -= dictationResultDelegate;
                _dRecognizer?.Stop();
                _dRecognizer?.Dispose();
                _dRecognizer = null;
                Log.d("sentence listener off"); 
            }
        }

        public void startSentenceRecognizer() {
            stopSentenceRecognizer();
            _dRecognizer = new DictationRecognizer();

            dictationResultDelegate = (text, confidence) => {
                Log.d($"Recognised phrase:{text}");
                if(currentSpeechCallback != null)
                    currentSpeechCallback(text);
            };

            _dRecognizer.DictationResult += dictationResultDelegate;

            try {
                _dRecognizer.Start();
                Log.d("sentence listener on");
            } catch (UnityEngine.UnityException ex) {
                Log.d("Problem with sentence listener");
            }
        }

        public void onSentenceSpoken(Action<string> action) {
            currentSpeechCallback = action;
        }

        private void OnPhraseRecognized(PhraseRecognizedEventArgs args){
			try{
                Log.d($"Recognised phrase:{args.text}");
				// if (_phrasePluginDict.ContainsKey (args.text)){
				// 	foreach (PluginReference pluginRef in _phrasePluginDict[args.text]){
				// 		Atom atom = SuperController.singleton.GetAtomByUid(pluginRef._atomName);JSONStorable pluginReceiver = atom.GetStorableByID(pluginRef._pluginName);JSONStorableStringChooser whenCondition = pluginReceiver.GetStringChooserJSONParam("whenCondition");JSONStorableBool speechHighConfidence = pluginReceiver.GetBoolJSONParam("SpeechHighConfidence");
				// 		if (whenCondition.val =="Speech Trigger"&& (!speechHighConfidence.val || args.confidence == ConfidenceLevel.High)){pluginReceiver.CallAction("Execute");}						
				// 	}
				// }
			} catch (Exception e) {SuperController.LogError("Exception caught: " + e);}
		}

        // public string DisplayMode(){
        //     var mode="VR";
        //     if (!SuperController.singleton.isOVR && !SuperController.singleton.isOpenVR){mode="2D";}
        //     else if(SuperController.singleton.mainHUD.Find("MainUICanvas/Panel/BackPanel").eulerAngles.x-SuperController.singleton.helpHUDText.transform.eulerAngles.x==0f){mode="2DVR";}
        //     return mode;
        // }

        // public bool realVR(){
        //     if(SuperController.singleton.mainHUD.Find("MainUICanvas/Panel/BackPanel").eulerAngles.x-SuperController.singleton.helpHUDText.transform.eulerAngles.x==0f){return false;}
        //     return true;
        // }


        private void ChangeHandState(string hand, int state)
        {
            switch(state)
            {
                case 0:
                    SingleFinger(hand, handStates[state]);
                    break;
                case 1:
                    SingleFinger(hand, handStates[state]);
                    break;
                case 2:
                    SingleFinger(hand, handStates[state]);
                    break;
                case 3:
                    SingleFinger(hand, handStates[state]);
                    break;
                case 4:
                    SingleFinger(hand, handStates[state]);
                    break;
                case 5:
                    ThumbIndex(hand);
                    break;
                case 6:
                    IndexMiddle(hand);
                    break;
                case 7:
                    IndexMiddleRing(hand);
                    break;
                case 8:
                    IndexMiddlePinky(hand);
                    break;
                case 9:
                    IndexRingPinky(hand);
                    break;
                case 10:
                    MakeFist(hand);
                    break;
            }
        }

        private void CloseFinger(string hand, string finger)
        {
            playerHands
                .GetComponentsInChildren<Transform>()
                    .Where(t => t.name.Contains($"{hand}_Finger_{finger}"))
                    .ForEach(t =>
                    {
                        float a = 3;
                        float b = 35;
                        float c = 35;

                        if(hand == "L")
                        {
                            a *= -1;
                            b *= -1;
                            c *= -1;
                        }

                        if(t.name.Contains("_A"))
                            t.Rotate(Vector3.forward, a);
                        else if (t.name.Contains("_B"))
                            t.Rotate(Vector3.forward, b);
                        else
                            t.Rotate(Vector3.forward, c);
                    });
        }

        private void MakeFist(string hand)
        {
            CloseFinger(hand, "Thumb");
            CloseFinger(hand, "Index");
            CloseFinger(hand, "Middle");
            CloseFinger(hand, "Ring");
            CloseFinger(hand, "Pinky");
        }

        private void SingleFinger(string hand, string finger)
        {
            foreach(var f in fingers)
            {
                if (f != finger)
                    CloseFinger(hand, f);
            }
        }

        private void ThumbIndex(string hand)
        {
            CloseFinger(hand, "Middle");
            CloseFinger(hand, "Ring");
            CloseFinger(hand, "Pinky");
        }

        private void IndexMiddle(string hand)
        {
            CloseFinger(hand, "Thumb");
            CloseFinger(hand, "Ring");
            CloseFinger(hand, "Pinky");

            //Experimental:  Currently has major collision issues and is basically not useable
            //playerHands
            //        .GetComponentsInChildren<Transform>()
            //        .Single(t => t.name.Contains($"{hand}_Finger_Middle_A"))
            //        .Rotate(Vector3.left, 4f);

            //playerHands
            //    .GetComponentsInChildren<Transform>()
            //    .Single(t => t.name.Contains($"{hand}_Finger_Index_A"))
            //    .Rotate(Vector3.right, 5f);
        }

        private void IndexMiddleRing(string hand)
        {
            CloseFinger(hand, "Thumb");
            CloseFinger(hand, "Pinky");
        }

        private void IndexMiddlePinky(string hand)
        {
            CloseFinger(hand, "Thumb");
            CloseFinger(hand, "Ring");
        }

        private void IndexRingPinky(string hand)
        {
            CloseFinger(hand, "Thumb");
            CloseFinger(hand, "Middle");
        }

        private void HandleUserInputStateChangeButton(ref bool stateChangeButtonPressed, ref int handStateIndex)
        {
            if(!stateChangeButtonPressed)
            {
                if (handStateIndex != handStates.Count - 1)
                {
                    handStateIndex++;
                }
                else
                {
                    handStateIndex = -1;
                }
            }
            
            stateChangeButtonPressed = true;
        }

        private void HandleUserInputStateResetButton(ref bool stateChangeButtonPressed, ref int handStateIndex)
        {
            if (!stateChangeButtonPressed)
            {
                handStateIndex = -1;
            }

            stateChangeButtonPressed = true;
        }






    }
}