using System;
using System.Collections.Generic;
using UnityEngine;

namespace Amaimon
{
    public class EventType
    {   // using enums crashes vam
        public const byte CARESS = 0;
        public const byte PAT = 1;
        public const byte HIT = 2;
    }

    public class CollisionEventArgs : EventArgs {
        public CollisionEventArgs(Collision c, byte t, AAtom o, GameObject op) {
            collision = c;
            evtType = t;
            offender = o;
            offendersPart = op;
        }

        public Collision collision { get;  }
        public byte evtType { get;  }
        public AAtom offender { get;  }
        public GameObject offendersPart { get;  }
    }




    public class TriggerEventArgs : EventArgs
    {
        public TriggerEventArgs(string trig, Collider colld) {
            trigger = trig;
            collider = colld;
        }
        public String trigger { get;  }
        public Collider collider { get;  }
    }

    class RegisterEntry {
        public float sumVelocity = 0;
        public float lastEnterTime = 0;
        public int count = 0;

        public float avgVelocity() {
            if(count == 0)
                return 0;
            else
                return sumVelocity / count;
        }
    }



    public class TriggerAnalyzer : MonoBehaviour {
        public event EventHandler<TriggerEventArgs> OnCollide;
        public string forTrigger = "";


        // collider ma attacherRigidbody
        // Colliders may be used as Triggers as well as blocking Rigidbodies. When a Collider is set as a Trigger, it can detect interactions with other game objects.
        //
        //When a GameObject collides with another GameObject, Unity calls OnTriggerEnter. - Collider:Component:UnityObject

        // OnCollisionEnter can be visualized as colliding against a wall - przedmiot się na tym zatrzyma
        // OnTriggerEnter can be visualized as triggering an alarm. - przedmiot przez to przejdzie

        /*
        OnTriggerEnter happens on the FixedUpdate function when two GameObjects collide. The Colliders involved are not always at the point of initial contact.
        Note: Both GameObjects must contain a Collider component. One must have Collider.isTrigger enabled, and contain a Rigidbody. 
        If both GameObjects have Collider.isTrigger enabled, no collision happens. The same applies when both GameObjects do not have a Rigidbody component.
        */
        private void OnTriggerEnter(Collider other)
        {
            // other.attachedRigidbody - co weszło w trigger
            //Log.d($"w OnTriggerEnter, attachedRb={other.attachedRigidbody}, gameObject={other.gameObject}");

            // w OnTriggerEnter, attachedRb=R_Finger_Index_A (UnityEngine.Rigidbody), gameObject=R_Finger_Index_A (UnityEngine.GameObject)
            // TRIG: CoreControl's R_Finger_Index_A  WITH Person's lNippleTrigger (CollisionTrigger)

            if(forTrigger == "") {
                // this happens when game object with this TouchAnalyzer enters ANY existing trigger
                //Log.d($"Warn - received trigger collision, but no trigger wa registereed, collider={other.name}, go={gameObject}");
            }

            OnCollide?.Invoke(this, new TriggerEventArgs(forTrigger, other));

            /*
                                // var shortestDistanceToLabia = colliders
                    //     .Select(col => Vector3.Distance(col.transform.position, abdomenControl.transform.position))
                    //     .DefaultIfEmpty(0)
                    //     .Min();
*/
        }

        //OnTriggerExit is called when the Collider other has stopped touching the trigger.
        /*
        OnTriggerExit is called when the Collider other has stopped touching the trigger.
        This message is sent to the trigger and the collider that touches the trigger. 
        Note: Trigger events are only sent if one of the colliders also has a rigidbody attached. 
        Trigger events will be sent to disabled MonoBehaviours, to allow enabling Behaviours in response to collisions.
        */
        private void OnTriggerExit(Collider other)
        {
            //OnCollideEvent(new TriggerEventArgs(null, EventType.CARESS, trigger, other));
        }
    }






    public class CollisionAnalyzer : MonoBehaviour
    {
        Dictionary<string, RegisterEntry> register = new Dictionary<string, RegisterEntry>();
        public event EventHandler<CollisionEventArgs> OnCollide;
        public AAtom offender;
        public GameObject offendersPart;

        static int instance;

        public CollisionAnalyzer() {
            instance++;
        }

        public void flush() {
            offender = null;
            offendersPart = null;
            OnCollide = null;
        }

        // Awake is called when the script instance is being loaded. - zamiast konstruktora
        // void Awake()
        // {

        // }


        // colllision ma w sobie collidera (that we hit), rigidbody (that we hit), gameobject (that we hit), relative velocity, transform udorzonego obiektu

        // OnCollisionEnter can be visualized as colliding against a wall
        // OnTriggerEnter can be visualized as triggering an alarm.

        // 	OnCollisionEnter is called when this collider/rigidbody has begun touching another rigidbody/collider.
        /*
        OnCollisionEnter is called when this collider/rigidbody has begun touching another rigidbody/collider.
        In contrast to OnTriggerEnter, OnCollisionEnter is passed the Collision class and not a Collider. 
        The Collision class contains information about contact points, impact velocity etc. 
        If you don't use collisionInfo in the function, leave out the collisionInfo parameter as this avoids unneccessary calculations. 
        Note: Collision events are only sent if one of the colliders also has a non-kinematic rigidbody attached. 
        Collision events will be sent to disabled MonoBehaviours, to allow enabling Behaviours in response to collisions.
        */
        protected virtual void OnCollisionEnter(Collision collision)
        {
            logCollision(collision, true);
        }

        // OnCollisionExit is called when this collider/rigidbody has stopped touching another rigidbody/collider.
        /*
        OnCollisionExit is called when this collider/rigidbody has stopped touching another rigidbody/collider.
        In contrast to OnTriggerExit, OnCollisionExit is passed the Collision class and not a Collider. 
        The Collision class contains information about contact points, impact velocity etc. 
        If you don't use collisionInfo in the function, leave out the collisionInfo parameter as this avoids unneccessary calculations. 
        Note: Collision events are only sent if one of the colliders also has a non-kinematic rigidbody attached. 
        Collision events will be sent to disabled MonoBehaviours, to allow enabling Behaviours in response to collisions.
        */
        protected virtual void OnCollisionExit(Collision collision)
        {
            logCollision(collision, false);
        }



        // szybkie, następujące po sobie (min 3 razy) na przemian enter/exit z małą szybkością = głaskanie
        // kilka różnych koliderów, mocno - uderzenie


        private void logCollision(Collision c, bool entering) {
            // mamy hashmapę [collision.collider] = collision.velocity (uśredniona), count, lastEnterTime
            // 

            //Log.d($"Velo={c.relativeVelocity.magnitude} {entering}");

            RegisterEntry entry;

            var now = Time.timeSinceLevelLoad;

            if(register.ContainsKey(c.collider.ToString())) {
                entry = register[c.collider.ToString()];
            }
            else {
                entry = new RegisterEntry();
                register[c.collider.ToString()] = entry;
            }

            if(entering) {
                // hitter started touching
                var velo = c.relativeVelocity.magnitude;
                if(entry.lastEnterTime == 0) {
                    //Log.d("Time was 0");
                    entry.lastEnterTime = now;
                    if(velo < 2f) {
                        OnCollide?.Invoke(this, new CollisionEventArgs(c, EventType.CARESS, offender, offendersPart));
                        //Log.d("Signalling caress");
                    }

                }
                // else
                //     Log.d($"{instance} Start of hit, lat={entry.lastEnterTime} {c.collider.ToString()}");

                entry.sumVelocity += velo;
                entry.count++;
            }
            else {
                // hitter stopped touching, is this the end of hitting?
                if(now - entry.lastEnterTime > 0.5) {
                    var velo = entry.avgVelocity();

//Log.d($"{instance} End of hit - velo = {velo} lat={entry.lastEnterTime} {c.collider.ToString()}");

                    if(velo < 2.5f) {
                        OnCollide?.Invoke(this, new CollisionEventArgs(c, EventType.CARESS, offender, offendersPart));
                    } 
                    else if(velo < 5f) {
                        OnCollide?.Invoke(this, new CollisionEventArgs(c, EventType.PAT, offender, offendersPart));
                    } 
                    else {
                        OnCollide?.Invoke(this, new CollisionEventArgs(c, EventType.HIT, offender, offendersPart));
                    }

                    entry.sumVelocity = 0;
                    entry.count = 0;
                    entry.lastEnterTime = 0;

                    //Log.d($"{instance} post reset - velo = {entry.sumVelocity} lat={entry.lastEnterTime} {c.collider.ToString()}");

                }
            }

            //Log.d($"logCollision {c.collider} count={entry.count} v={entry.avgVelocity}");
        }
    }
}
