using System;
using System.Collections.Generic;
using UnityEngine;
using System.Linq;
using MVR.FileManagementSecure;
using MVR.FileManagement;

namespace Amaimon {
    public static class GenericExtensions {
        public static T Apply<T>(this T thiz, Action<T> a) {
            a(thiz);
            return thiz;
        }
    }

    public static class QuaternionExtensions
    {
        // Source: https://gist.github.com/maxattack/4c7b4de00f5c1b95a33b
        public static Quaternion SmoothDamp(this Quaternion current, Quaternion target, ref Quaternion currentVelocity, float smoothTime)
        {
            // account for double-cover
            var dot = Quaternion.Dot(current, target);
            var multi = dot > 0f ? 1f : -1f;
            target.x *= multi;
            target.y *= multi;
            target.z *= multi;
            target.w *= multi;
            // smooth damp (nlerp approx)
            var result = new Vector4(
                Mathf.SmoothDamp(current.x, target.x, ref currentVelocity.x, smoothTime),
                Mathf.SmoothDamp(current.y, target.y, ref currentVelocity.y, smoothTime),
                Mathf.SmoothDamp(current.z, target.z, ref currentVelocity.z, smoothTime),
                Mathf.SmoothDamp(current.w, target.w, ref currentVelocity.w, smoothTime)
            ).normalized;
            // compute deriv
            var dtInv = 1f / Time.smoothDeltaTime;
            currentVelocity.x = (result.x - current.x) * dtInv;
            currentVelocity.y = (result.y - current.y) * dtInv;
            currentVelocity.z = (result.z - current.z) * dtInv;
            currentVelocity.w = (result.w - current.w) * dtInv;
            return new Quaternion(result.x, result.y, result.z, result.w);
        }


    }

    public static class JSONStorabeleEXtensions {
    }

    public static class AtomExtensions {
        public static void callAction(this Atom person, string actionName, string argsName, Action<JSONStorable> argsFunction) {
            JSONStorable js = person.GetStorableByID(argsName);
            argsFunction(js);
            js.CallAction(actionName);
        }

        public static void parentLinkAtPosition(this Atom a, Atom parent, JSONStorable toPosition, string toPositionName)
        {
            a.parentAtom = parent;
            Vector3 p = toPosition.transform.position;
            a.mainController.currentPositionState = FreeControllerV3.PositionState.On;
            a.mainController.transform.position = new Vector3(p.x, p.y, p.z);
            a.mainController.SetLinkToAtom(parent.uid);
            a.mainController.SetLinkToRigidbodyObject(toPositionName);
            a.mainController.currentPositionState = FreeControllerV3.PositionState.ParentLink;
            a.mainController.canGrabPosition = false;
            a.mainController.canGrabRotation = false;
        }

        public static T getParam<T>(this Atom atom, string receiverName, string paramName) where T : JSONStorableParam {
            JSONStorable receiver = atom.GetStorableByID(receiverName);

            if(typeof(T) == typeof(JSONStorableColor))
                return receiver.GetColorJSONParam(paramName) as T;
            if(typeof(T) == typeof(JSONStorableFloat))
                return receiver.GetFloatJSONParam(paramName) as T;
            if(typeof(T) == typeof(JSONStorableBool))
                return receiver.GetBoolJSONParam(paramName) as T;
            if(typeof(T) == typeof(JSONStorableStringChooser))
                return receiver.GetStringChooserJSONParam(paramName) as T;
            else
                throw new Exception("Bad type for getParam"); 
        }


        public static void dumpNodes(this Atom person) {
            if(person != null) {
                // foreach (FreeControllerV3 fc in person.freeControllers){
                //     Log.d($"{fc?.name}");
                // }

                foreach (Rigidbody fc in person.rigidbodies){
                    Log.d($"{fc?.name}");
                }
            }
        }



        public static void dumpJointState(this Atom person) {
            foreach(FreeControllerV3 fc in person.freeControllers) {
                var positionState = fc.currentPositionState;
                if(positionState != null) {
                    Log.d($"{fc} = {positionState}");
                }
            }
        }
    }

    public static class RigidbodyExtensions {
        public static string niceName(this Rigidbody rb) {
            return rb.name.ToLower();
        }
    }

    public static class FreeControllerExtensions {
        public static string niceName(this FreeControllerV3 c) {
            return c.name.ToLower();
        }
    }

    public static class LinqExtensions {
        public static void ForEach<T>(this IEnumerable<T> source, Action<T> action)
        {
            if (source == null) throw new ArgumentNullException("source");
            if (action == null) throw new ArgumentNullException("action");

            foreach (T item in source)
            {
                action(item);
            }
        }
    }

    public static class TransformExtensions {
        public static void dumpTree(this Transform go, List<string> FileLines, int depth = 0) {
            for(int i = 0; i<go.childCount; i++) {
                var child =  go.GetChild(i);
                var wciecie = new string(' ', depth);
                FileLines.Add(wciecie+child);
                if(child.childCount > 0) {
                    dumpTree(child, FileLines, depth+1);
                }
            }
        }

        // public static void textfind(this Transform t, string str=""){
        //     foreach(Text im in t.GetComponentsInChildren<Text>(true)){
        //         if(im != null && im.text !=null && im.text.Contains(str)){
        //             string s = ""; 
        //             Transform g = im.transform.parent;int i=0;
        //             while (g != null && i<3)
        //             {
        //                 i++;
        //                 s = g.name + "/" + s;
        //                 g = g.parent;
        //             }
        //             var find="\""+s+"\"";
        //             SuperController.LogMessage("Examples:\n"+t.name+".transform.Find("+find+").GetComponentInChildren<Button>(true).onClick.Invoke();\n"+t.name+".transform.Find("+find+").GetComponentInChildren<UIDynamicToggle>(true).toggle=true;\n"+t.name+".transform.Find("+find+").GetComponentInChildren<Image>(true).color=new Color(0.5f,0.5f,0.5f,0.5f);\n\n");
        //         }
        //     }
        // }

    }

    public static class FileWriter {
        private static string _plugindir = "Saves\\PluginData";
        private static string _pluginowner = "\\Amaimon";
        private static string _plugin = "\\Story";
        public static void write(List<string> FileLines,string name) {
            plugindir();
            FileManagerSecure.WriteAllText(_plugindir+_pluginowner+_plugin+name, String.Join("\n", FileLines.ToArray()));
        }

        public static bool plugindir(){
            bool result = false;
            if (FileManagerSecure.DirectoryExists(_plugindir))
            {    
                if (!FileManagerSecure.DirectoryExists(_plugindir+_pluginowner))
                FileManagerSecure.CreateDirectory(_plugindir+_pluginowner);

                if (FileManagerSecure.DirectoryExists(_plugindir+_pluginowner)){
                    if (!FileManagerSecure.DirectoryExists(_plugindir+_pluginowner+_plugin))
                    FileManagerSecure.CreateDirectory(_plugindir+_pluginowner+_plugin);
                }             
            }
            else Log.d("This function requires the folder Saves\\PluginData to exist.");
            return (result);
        }
    }

}