using System;
using UnityEngine;

namespace Amaimon {

        public class BreathingInterpolator : OscillatingInterpolable {
            DAZMorph chest;
            DAZMorph stomach;
            DAZMorph noseIn;
            DAZMorph noseOut;
            FreeControllerV3 chestController;
            FreeControllerV3 rShoulderController;
            FreeControllerV3 lShoulderController;

            public BreathingInterpolator(AActor person, float perdiodSecs, float sustainCycles = 100000) : base(perdiodSecs, sustainCycles) {
                _phase = 1;

                chest = person.getMorph("Breathing Chest");
                noseIn = person.getMorph("Breathing NoseIn");
                noseOut = person.getMorph("Breathing NoseOut");
                stomach = person.getMorph("Breathing Stomach");

                chestController = person.getController("chestControl");
                rShoulderController = person.getController("rShoulderControl");
                lShoulderController = person.getController("lShoulderControl");

                // if(_morph == null)
                //     Log.d($"Couldn't find morph {name} on {person.person.name}");
                // else
                //     _default = _morph.morphValue;

            }

        override public void onPhase(float phase) 
        {

            //Log.d($"Breath phase {phase}");

            var rad =  phase * Mathf.Deg2Rad;

            var periodCosinus = (float)(System.Math.Cos(rad));
 
            var chestValue = (-periodCosinus / 2) + 0.5f; // values from 0 to 1
 
            chestController.jointRotationDriveSpring = 1000;
            chestController.jointRotationDriveDamper = 235;
            chestController.jointRotationDriveXTarget = -periodCosinus * 4.3f;

            rShoulderController.jointRotationDriveSpring = 10 + periodCosinus * 10;
            rShoulderController.jointRotationDriveDamper = 400;
            lShoulderController.jointRotationDriveSpring = 10 + periodCosinus * 10;
            lShoulderController.jointRotationDriveDamper = 400;

            rShoulderController.jointRotationDriveXTarget = -periodCosinus * 50f;
            rShoulderController.jointRotationDriveYTarget = periodCosinus * 50f;

            lShoulderController.jointRotationDriveXTarget = periodCosinus * 50f;
            lShoulderController.jointRotationDriveYTarget = -periodCosinus * 50f;

            chest.morphValue = chestValue;  
            noseOut.morphValue = 1f - chestValue / 1f;
            stomach.morphValue = chestValue / 2f;
        }
    }
}