using System;
using System.Linq;

using UnityEngine;
using UnityEngine.Events;
using UnityEngine.UI;
using UnityEngine.XR;

using SimpleJSON;

namespace Amaimon {

    // reload: onDisable -> onDestroy -> onEnable -> Init
    // disable: OnDisable
    // enable: OnEnable
    // wczytanie w nowej scenie: onEnable -> Init

    // OnEnable -> [Init] -> OnDisable -> OnD-estroy

    public class DancingFire : MVRScript
    {
        JSONStorableString _debug;		
        protected JSONStorableFloat _slider1;
        protected JSONStorableFloat _slider2;
        protected JSONStorableFloat _slider3;
        protected JSONStorableFloat _slider4;

        UIDynamicTextField _debugTextField;

        JSONStorable receiver;
        JSONStorableFloat intensity;
        float baseIntensity;
        Vector3 basePosition;

        FreeControllerV3 control;

        float nextFrame;

        float movementRange = 0.50f;

        float ix =0;
        float iy=0;

        public Vector3 position {
            get {
                return control.transform.position;
            }
            set {
                control.transform.position = value;
                //person.ResetPhysics(true, true);
            }
        }


        ///////////////////////////////////////////////////
        // Update()
        // Run things every frame
        ///////////////////////////////////////////////////

        public void Update()
        {
            // EVERY FRAME
            // containingAtom
            // Atom atom = SuperController.singleton.GetAtomByUid("jakieś światło");

            if(nextFrame <= 0) {
                var ychange = Mathf.PerlinNoise(0, iy) * movementRange;
                var xchange = Mathf.PerlinNoise(ix, 0) * movementRange;
                var yshift = Vector3.up * ychange;
                var xshift = Vector3.right * xchange;

                iy+=0.1f;
                ix+=0.1f;

                position = basePosition + yshift + xshift;

                intensity.val = baseIntensity + (Mathf.PerlinNoise(0, iy)-0.5f) * 1.20f;
                //Log.d($"intensity={intensity.val}");

                nextFrame = UnityEngine.Random.Range(0.005f, 0.02f);
            }
            else
                nextFrame -= Time.deltaTime;
        }

 
         ///////////////////////////////////////////////////
        // OnEnable()
        // When plugin enabled
        ///////////////////////////////////////////////////
        public void OnEnable()
        {

        }


        ///////////////////////////////////////////////////
        // Init()
        // Run once, when plugin loaded
        ///////////////////////////////////////////////////
        public override void Init()
        {
            // called only once!
            // here GUI buttons migt be set up

            try
            {
                _debug = new JSONStorableString("PluginDebug","");
                _debugTextField = CreateTextField(_debug,false);
                _debugTextField.height = 20;		

                // default, min, max

                _slider2 = new JSONStorableFloat("unused", 0f, -1.0f, 1f, true, true);
                RegisterFloat(_slider2);
                CreateSlider(_slider2,false);

                _slider3 = new JSONStorableFloat("unused", 0f, -1.0f, 1f, true, true);
                RegisterFloat(_slider3);
                CreateSlider(_slider3,false);

                _slider4 = new JSONStorableFloat("unused", 0f, 0f, 1f, true, true);
                RegisterFloat(_slider4);
                CreateSlider(_slider4,false);

                _slider1 = new JSONStorableFloat("unused", 0.0f, -1.0f, 1.0f, true, true);
                RegisterFloat(_slider1);
                CreateSlider(_slider1,false);

                receiver = containingAtom.GetStorableByID("Light");
                intensity = receiver.GetFloatJSONParam("intensity"); 

                // var all = receiver.GetFloatParamNames();
                // Log.d($"all={all}");
                // all.ForEach(it => Log.d($"recv={it}"));

                control = containingAtom.freeControllers.FirstOrDefault(fc => fc.name == "control");
                baseIntensity = intensity.val;
                basePosition = position;
                //Log.d($"recv={receiver} int={intensity}, control={control}");
            }
            catch (Exception e)
            {
                //SuperController.LogError($"{nameof(StoryPlugin)}.{nameof(Init)}: {e}");
            }
        }

        ///////////////////////////////////////////////////
        // OnDisable()
        // Run things when plugin disabled
        ///////////////////////////////////////////////////
        public void OnDisable()
        {
            try
            {
                intensity.val = baseIntensity;
                position = basePosition;
            }
            catch (Exception e)
            {
                //SuperController.LogError($"{nameof(StoryPlugin)}.{nameof(OnDisable)}: {e}");
            }
        }

        ///////////////////////////////////////////////////
        // OnDestroy()
        // Run when plugin removed
        ///////////////////////////////////////////////////
        public void OnDestroy()
        {
            try
            {

            }
            catch (Exception e)
            {
                //SuperController.LogError($"{nameof(StoryPlugin)}.{nameof(OnDestroy)}: {e}");
            }
        }

        ///////////////////////////////////////////////////
        // Various utility functions
        ///////////////////////////////////////////////////
    }
}
