using System;
using System.Linq;
using System.Collections;
using System.Collections.Generic;

using UnityEngine;

namespace Amaimon {
    public class StdDevTracker {
        int sampleCount;
        float sum;
        float sqsum;

        public void register(float sample) {
            sum += sample;
            sqsum += sample * sample;
            sampleCount++;
        }

        virtual public float get() {
            if(sampleCount<2)
                return 0;
            
            var avg = sum/sampleCount;
            return (float)Math.Sqrt(sampleCount/(sampleCount-1)*(sqsum/sampleCount-(avg*avg)));
        }

        public void reset() {
            sampleCount = 0;
            sum = 0;
            sqsum = 0;
        }
    }

    public class RhytmTracker : StdDevTracker {

        float prevTime = -1;
        public bool isRythmic = false;

        public void click() {
            if(prevTime == -1) {
                prevTime = Time.unscaledTime;
            }
            else {
                var dt = Time.unscaledTime - prevTime;
                if(dt > 3f) {
                    Log.d("Ryhtm too slow, resetting");
                    reset();
                    prevTime = Time.unscaledTime;
                    isRythmic = false;
                }
                else {
                    prevTime = Time.unscaledTime;
                    var prevValue = get();
                    register(dt);
                    var newValue = get();

                    var delta = prevValue - newValue;
                    // > 0 - rythm improving
                    // < 0 - rythm worsening
                    isRythmic = delta > 0;

                    Log.d($"isRythmic = {isRythmic}, {1f/dt}/s");
                }
            }
        }
   }
}